// xcc xCode C compiler
// Copyright Adam Chlipala 1999
// 23 skidoo

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#include <math.h>

#ifdef WIN32
	#include <direct.h>
#endif

#ifdef WIN32
	#define _int64 __int64
	#define _float64 long double
#else
	#define _int64 long long
	#define _float64 double
#endif

char *_strrev(char *str)
{
	char *str2 = malloc(strlen(str)+1), *sp, *sp2;

	strcpy(str2, str);
	sp = str;
	sp2 = &str2[strlen(str2)-1];

	for (; sp2 >= str2; --sp2)
		*sp++ = *sp2;
	*sp = '\0';
	return str;
}

int mantissa[4] = {0, 0, 23, 52};
int exponent[4] = {0, 0, 8, 11};
int bias[4] = {0, 0, 127, 1023};
char tsz[4] = {1, 2, 4, 8};

#define PATH_LENGTH 255
#define BLOCK_SIZE 1024
#define WORD_LENGTH 1024

#define LAST_ASM (asm_code->cur_asm->str)

#define LOG_10_2 0.301029995664

typedef struct lstr
{
	struct lstr *next, *prev;
	char *str, *rep;
	unsigned size:4, flag:1;
} lstr;

typedef struct line_info
{
	int line, pos;
	void *file;
} line_info;

typedef struct file_info
{
	struct file_info *next, *prev;
	char *name, *text;
	line_info *lines;
	int n_lines, size;
} file_info;

typedef struct macro
{
	struct macro *next, *prev;
	char *name, *value;
	lstr *args;
} macro;

typedef struct pevent
{
	struct pevent *next, *prev;
	char str[255];
	int pos;
} pevent;

pevent *pevent_base = NULL, *cur_pevent = NULL;

#define N_ATTRS 24
char *attrs[N_ATTRS] = {"auto", "const", "extern", "register",
	"signed", "static", "unsigned", "typedef", "volatile",
	"void", "char", "int", "short", "long", "float", "double",
	"__host__", "__input__", "__output__", "__string__", "__sync__",
	"struct", "union", "enum"};
#define AUTO (1<<0)
#define CONST (1<<1)
#define EXTERN (1<<2)
#define REGISTER (1<<3)
#define SIGNED (1<<4)
#define STATIC (1<<5)
#define UNSIGNED (1<<6)
#define TYPEDEF (1<<7)
#define VOLATILE (1<<8)
#define VOID (1<<9)
#define CHAR (1<<10)
#define INT (1<<11)
#define SHORT (1<<12)
#define LONG (1<<13)
#define FLOAT (1<<14)
#define DOUBLE (1<<15)
#define HOST (1<<16)
#define INPUT (1<<17)
#define OUTPUT (1<<18)
#define STRING (1<<19)
#define SYNC (1<<20)
#define STRUCT (1<<21)
#define UNION (1<<22)
#define ENUM (1<<23)
#define NEEDS_ARG(X) ((X) >= 21)
#define TYPECMP (SIGNED | UNSIGNED | VOID | CHAR | INT | SHORT | LONG | FLOAT | DOUBLE | STRUCT | UNION | ENUM)
#define XCATTRS (HOST | INPUT | OUTPUT | STRING | SYNC)

#define N_KEYWORDS 37
char *keywords[N_KEYWORDS] = {"auto", "const", "extern", "register",
	"signed", "static", "unsigned", "typedef", "volatile",
	"void", "char", "int", "short", "long", "float", "double",
	"struct", "union", "enum", "if", "else", "while", "do",
	"for", "switch", "case", "default", "goto", "break", "continue",
	"return", "sizeof", "__host__", "__input__", "__output__",
	"__string__", "__sync__"};

typedef struct lint
{
	struct lint *prev, *next;
	int n;
} lint;

typedef struct econst
{
	struct econst *next;
	char *name;
	int val;
} econst;

typedef struct enumt
{
	struct enumt *next;
	char *name;
	econst *econst_base;
} enumt;

struct var;

typedef struct stun
{
	struct stun *next;
	char *name, defd;
	struct var *var_base, *cur_var;
	int size;
} stun;

typedef union tag
{
	enumt *e;
	stun *s;
} tag;

typedef struct var
{
	struct var *next, *prev, *next_arg, *prev_arg, *args;
	char *name, *init_val, type;
	unsigned int global:1, func:1, fptr:1, argarr:1, vararg:1;
	int loc, ptr, size, unit_size, load_size, host, nargs;
	unsigned int attrs;
	lint *arr_base;
	tag t;
	char bf_loc, bf_size;
} var;

typedef struct oper
{
	char *op, *cmd;
	int type, order;
} oper;

typedef struct asmt
{
	struct asmt *next, *prev;
	char *str;
} asmt;

#define DECLR 1
#define CMND 2
#define MULTI 3
#define FUNC 4
typedef struct code
{
	struct code *prev, *next, *peer, *child, *parent, *cparent;
	struct code *pparent;
	char *str, type, *_break, *_continue, *exprq;
	file_info *file;
	int pos, size, self_size, child_size;
	var *var_base, *td_base;
	asmt *asm_base, *cur_asm, *aq, *aq_base;
	lstr *label_base, *forced_label_base;
	enumt *enum_base;
	stun *struct_base, *union_base;
} code;

typedef struct ulbl
{
	struct ulbl *next, *prev;
	char *user, *real;
	code *c;
	int pos;
} ulbl;

typedef struct func
{
	struct func *prev, *next;
	var type;
	code *code_base;
	ulbl *label_base, *orphan_base;
} func;

typedef struct caset
{
	struct caset *next, *prev;
	char *label;
	int val;
} caset;

#define BINARY 1
#define PREFIX 2
#define POSTFIX 3
#define TERNARY 4

#define N_OPERS 44
#define N_ORDERS 15
oper opers[N_OPERS] =
{
	{".", "DOT", 1, 1},
#define DOT 0
	{"->", "ARROW", 1, 1},
#define ARROW 1
	{"++", "INC_POST", 3, 1},
#define INC_POST 2
	{"--", "DEC_POST", 3, 1},
#define DEC_POST 3
	{"++", "INC_PRE", 2, 2},
#define INC_PRE 4
	{"--", "DEC_PRE", 2, 2},
#define DEC_PRE 5
	{"*", "INDIR", 2, 2},
#define INDIR 6
	{"&", "ADDR", 2, 2},
#define ADDR 7
	{"+", "PLUS", 2, 2},
#define PLUS 8
	{"-", "NEG", 2, 2},
#define NEG 9
	{"!", "BNOT", 2, 2},
#define BNOT 10
	{"~", "NOT", 2, 2},
#define NOT 11
	{"*", "MULT", 1, 3},
#define MULT 12
	{"/", "DIV", 1, 3},
#define DIV 13
	{"%", "MOD", 1, 3},
#define MOD 14
	{"+", "ADD", 1, 4},
#define ADD 15
	{"-", "SUB", 1, 4},
#define SUB 16
	{"<<", "SLN", 1, 5},
#define SLN 17
	{">>", "SRN", 1, 5},
#define SRN 18
	{"<", "LT", 1, 6},
#define LT 19
	{">", "GT", 1, 6},
#define GT 20
	{"<=", "LTE", 1, 6},
#define LTE 21
	{">=", "GTE", 1, 6},
#define GTE 22
	{"==", "EQ", 1, 7},
#define EQ 23
	{"!=", "NEQ", 1, 7},
#define NEQ 24
	{"&", "AND", 1, 8},
#define AND 25
	{"^", "XOR", 1, 9},
#define XOR 26
	{"|", "OR", 1, 10},
#define OR 27
	{"&&", "BAND", 1, 11},
#define BAND 28
	{"||", "BOR", 1, 12},
#define BOR 29
	{"?", "QMARK", 4, 13},
#define QMARK 30
	{":", "COLON", 4, 13},
#define COLON 31
	{"=", "ASN", 1, 14},
#define ASN 32
	{"*=", "ASN_MULT", 1, 14},
#define ASN_MULT 33
	{"/=", "ASN_DIV", 1, 14},
#define ASN_DIV 34
	{"%=", "ASN_MOD", 1, 14},
#define ASN_MOD 35
	{"+=", "ASN_ADD", 1, 14},
#define ASN_ADD 36
	{"-=", "ASN_SUB", 1, 14},
#define ASN_SUB 37
	{"<<=", "ASN_SLN", 1, 14},
#define ASN_SLN 38
	{">>=", "ASN_SRN", 1, 14},
#define ASN_SRN 39
	{"&=", "ASN_AND", 1, 14},
#define ASN_AND 40
	{"^=", "ASN_XOR", 1, 14},
#define ASN_XOR 41
	{"|=", "ASN_OR", 1, 14},
#define ASN_OR 42
	{",", "COMMA", 1, 15}
#define COMMA 43
};
#define LVAL(X) (((X) >= ASN && (X) <= ASN_OR) || \
	((X) >= INC_POST && (X) <= DEC_PRE) || (X) == ADDR)
#define IS_ASN(X) ((X) == 14)
#define IS_TEST(X) ((X) == 6 || (X) == 7)

file_info finfo, finfo2, *flist_base = NULL;
macro *macro_base = NULL;
int errcount = 0, warncount = 0;
char date_str[100], time_str[100];
file_info *cur_file = NULL;
int cur_line = -1;
func global, *func_base = NULL, *cur_func;
int line_pos = 0, last_line_pos = 0;
code *asm_code;
int n_labels = 0, host_funcs = 0, host_vars = 0, cur_data = 0;
lstr *label_base = NULL, *temp_td_base = NULL;
char muzzle = 0;
int inc_level = 0;

int parse_declr(char *, var *, code *, stun *, char);

char *op_chars = "+-=&|<>*!/%^~,?:.";
char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
"Aug", "Sep", "Oct", "Nov", "Dec"};

#define _C_SIGNED 1
#define _C_UNSIGNED 2
#define _C_FLOAT 3

#define INC_MAX 20

struct
{
	unsigned int ansi:1, ansi_on:1, skip_asm:1, save_asm:1, nonet:1,
		host_table:1;
} settings;

char path[PATH_LENGTH];

_int64 atoi64(char *str)
{
	_int64 ret = 0, sign = 1;

	if (*str == '-')
	{
		sign = -1;
		++str;
	}
	for (; *str; ++str)
	{
		ret *= 10;
		ret += *str - '0';
	}
	return ret * sign;
}

int is_opchar(char ch)
{
	char *p;

	for (p = op_chars; *p; ++p)
		if (ch == *p)
			return 1;
	return 0;
}

int is_valid_id(char *str)
{
	int j;
	char *sp = str;

	if (!isalpha(*sp) && *sp != '_')
		return 0;
	for (; *sp; ++sp)
		if (!isalnum(*sp) && *sp != '_')
			return 0;

	for (j = 0; j < N_KEYWORDS; ++j)
		if (!strcmp(str, keywords[j]))
			return 0;

	return 1;
}

int id_free(char *str, code *c, func *f)
{
	var *v;
	ulbl *ls;
	enumt *e;
	econst *ec;

	if (!c)
		c = global.code_base;
	for (v = c->var_base; v; v = v->next)
		if (!strcmp(v->name, str))
			return 0;
	for (v = c->td_base; v; v = v->next)
		if (!strcmp(v->name, str))
			return 0;
	for (e = c->enum_base; e; e = e->next)
	{
/*		if (e->name)
			if (!strcmp(e->name, str))
				return 0;
*/		for (ec = e->econst_base; ec; ec = ec->next)
			if (!strcmp(ec->name, str))
				return 0;
	}
	if (f)
		for (ls = f->label_base; ls; ls = ls->next)
			if (!strcmp(ls->user, str))
				return 0;

	return 1;
}

var *find_var(char *str, code *c)
{
	var *v;

	if (!c)
		return NULL;
	for (v = c->var_base; v; v = v->next)
		if (!strcmp(v->name, str))
			return v;

	if (!v && c->parent)
		return find_var(str, c->parent);
	return NULL;
}

enumt *find_enum(char *str, code *c)
{
	enumt *e;

	if (!c)
		return NULL;
	for (e = c->enum_base; e; e = e->next)
		if (e->name)
			if (!strcmp(e->name, str))
				return e;

	if (c->parent)
		return find_enum(str, c->parent);
	return NULL;
}

econst *find_econst(char *str, code *c, enumt **type)
{
	enumt *e;
	econst *ec;

	if (!c)
		return NULL;
	for (e = c->enum_base; e; e = e->next)
		for (ec = e->econst_base; ec; ec = ec->next)
			if (!strcmp(ec->name, str))
			{
				if (type)
					*type = e;
				return ec;
			}

	if (c->parent)
		return find_econst(str, c->parent, type);
	return NULL;
}

stun *find_struct(char *str, code *c)
{
	stun *s;

	if (!c)
		return NULL;
	for (s = c->struct_base; s; s = s->next)
		if (s->name)
			if (!strcmp(s->name, str))
				return s;

	if (c->parent)
		return find_struct(str, c->parent);
	return NULL;
}

stun *find_union(char *str, code *c)
{
	stun *s;

	if (!c)
		return NULL;
	for (s = c->union_base; s; s = s->next)
		if (s->name)
			if (!strcmp(s->name, str))
				return s;

	if (c->parent)
		return find_union(str, c->parent);
	return NULL;
}

var *find_stun_member(char *str, stun *s)
{
	var *v;

	for (v = s->var_base; v; v = v->next)
		if (v->name)
			if (!strcmp(str, v->name))
				return v;

	return NULL;
}

var *find_td(char *str, code *c)
{
	var *v;

	if (!c)
		return NULL;
	for (v = c->td_base; v; v = v->next)
		if (!strcmp(v->name, str))
			return v;

	if (c->parent)
		return find_td(str, c->parent);
	return NULL;
}

func *find_func(char *str)
{
	func *f;

	for (f = func_base; f; f = f->next)
		if (!strcmp(f->type.name, str))
			return f;

	return NULL;
}

int is_const(char *str, code *c)
{
	int j, paren = 0;
	char sword[WORD_LENGTH], *sp = sword, esc = 0;

	for (; ; ++str)
	{
		if (*str == '"')
			return 0;
		if (*str == '\'')
		{
			for (++str; *str; ++str)
			{
				if (esc)
					esc = 0;
				else if (*str == '\\')
					esc = 1;
				else if (*str == '\'')
					break;
			}
		}
		else if (*str != '_' && !isalnum(*str))
		{
			*sp = '\0';
			if (sword[0])
			{
				if (find_econst(sword, c, NULL));
				else if (!strcmp(sword, "sizeof"))
				{
					for (; *str == ' '; ++str);
					for (; *str; ++str)
					{
						if (paren)
						{
							if (*str == '(' || *str == '[')
								++paren;
							else if (*str == ')' || *str == ']')
								if (!--paren)
									break;
						}
						else if (*str == '(')
							paren = 1;
						else if (*str != '_' && !isalnum(*str))
						{
							--str;
							break;
						}
					}
				}
				else if (!isdigit(sword[0]))
					return 0;
			}
			sp = sword;
		}
		else
			*sp++ = *str;
		if (!*str)
			break;
	}

	return 1;
}

char *trim_spaces(char *str)
{
	int sz, j, k;

	if (!str) return NULL;
	sz = strlen(str);
	for (j = 0; j < sz; ++j)
	{
		if (!isspace(str[j]))
		{
			if (j)
				for (k = j; k <= sz; ++k)
					str[k-j] = str[k];
			break;
		}
	}

	for (j = strlen(str)-1; j >= 0; --j)
	{
		if (!isspace(str[j]))
		{
			str[j+1] = '\0';
			break;
		}
	}

	return str;
}

char *first_word(char *buf, char *src)
{
	char *sp = buf;

	for (; *src; ++src)
		if (!isalpha(*src) && !isdigit(*src))
			break;
		else
			*sp++ = *src;
	*sp = '\0';
	return buf;
}

char *standardize_path(char *str)
{
	char *ptr;

	for (ptr = str; *ptr; ++ptr)
		if (*ptr == '\\')
			*ptr = '/';
	return str;
}

void free_lstr(lstr *ptr)
{
	lstr *ptr2;

	for (; ptr; ptr = ptr2)
	{
		ptr2 = ptr->next;
		free(ptr->str);
		if (ptr->rep)
			free(ptr->rep);
		free(ptr);
	}
}

void free_lstr2(lstr *ptr)
{
	lstr *ptr2;

	for (; ptr; ptr = ptr2)
	{
		ptr2 = ptr->next;
		free(ptr);
	}
}

void free_lint(lint *lnt)
{
	lint *lnt2;

	for (; lnt; lnt = lnt2)
	{
		lnt2 = lnt->next;
		free(lnt);
	}
}

void free_ulbl(ulbl *ptr)
{
	ulbl *ptr2;

	for (; ptr; ptr = ptr2)
	{
		ptr2 = ptr->next;
		free(ptr->user);
		free(ptr);
	}
}

void free_var(var *v, char fr)
{
	var *vp, *vp2;

	if (!v)
		return;
	if (v->name)
		free(v->name);
	if (v->init_val)
		free(v->init_val);
	if (v->arr_base)
		free_lint(v->arr_base);

	for (vp = v->args; vp; vp = vp2)
	{
		vp2 = vp->next_arg;
		free_var(vp, 1);
	}

	if (fr)
		free(v);
}

void free_enum(enumt *e)
{
	enumt *e2;
	econst *ec, *ec2;

	for (; e; e = e2)
	{
		e2 = e->next;
		if (e->name)
			free(e->name);
		for (ec = e->econst_base; ec; ec = ec2)
		{
			ec2 = ec->next;
			if (ec->name)
				free(ec->name);
			free(ec);
		}
		free(e);
	}
}

void free_stun(stun *s)
{
	stun *s2;

	for (; s; s = s2)
	{
		s2 = s->next;
		if (s->name)
			free(s->name);
		free_var(s->var_base, 1);
		free(s);
	}
}

void free_code(code *c)
{
	code *c2;
	var *v, *v2;
	asmt *a, *a2;

	if (c->exprq)
		free(c->exprq);
	if (c->label_base)
		free_lstr(c->label_base);

	for (a = c->asm_base; a; a = a2)
	{
		a2 = a->next;
		free(a->str);
		free(a);
	}

	for (v = c->var_base; v; v = v2)
	{
		v2 = v->next;
		free_var(v, 1);
	}
	for (v = c->td_base; v; v = v2)
	{
		v2 = v->next;
		free_var(v, 1);
	}
	if (c->enum_base)
		free_enum(c->enum_base);
	if (c->struct_base)
		free_stun(c->struct_base);
	if (c->union_base)
		free_stun(c->union_base);

	for (; c; c = c2)
	{
		c2 = c->next;
		if (c->child)
			free_code(c->child);
		if (c->peer)
			free_code(c->peer);
		if (c->str)
			free(c->str);
	}
}

int typecmp(var *v1, var *v2)
{
	var *a1, *a2;
	lint *n1, *n2;

	if ((v1->attrs & TYPECMP) != (v2->attrs & TYPECMP) ||
		v1->ptr != v2->ptr || v1->func != v2->func)
		return 1;

	if ((v1->attrs & (STRUCT | UNION | ENUM)) &&
		(v2->attrs & (STRUCT | UNION | ENUM)) &&
		v1->t.s != v2->t.s)
		return 1;

	for (a1 = v1->args, a2 = v2->args; a1; a1 = a1->next_arg,
		a2 = a2->next_arg)
	{
		if (!a2)
			return 1;
		if (typecmp(a1, a2))
			return 1;
	}
	if (a2)
		return 1;

	for (n1 = v1->arr_base, n2 = v2->arr_base; n1; n1 = n1->next,
		n2 = n2->next)
	{
		if (!n2)
			return 1;
		if (n1->n != n2->n)
			return 1;
	}
	if (n2)
		return 1;

	return 0;
}

void pack_int(char *outp, char *str, char len)
{

	unsigned _int64 i;
	char j;

	if (*str == '-')
		i = pow(2, 8*len) - atoi64(&str[1]);
	else
		i = atoi64(str);
	for (j = 0; j < len; ++j)
	{
		*outp++ = i & 255;
		i >>= 8;
	}

}

int log2(unsigned _int64 i)
{
#ifdef WIN32
	unsigned _int64 cb = 0x8000000000000000;
#else
	unsigned _int64 cb = 0x8000000000000000ULL;
#endif
	int j;

	for (j = 63; j >= 0; --j)
	{
		if (i & cb)
			return j;
		cb >>= 1;
	}
	return 0;
}

void pack_int2(char *outp, unsigned _int64 i, char len)
{
	char j;

	for (j = 0; j < len; ++j)
	{
		*outp++ = i & 255;
		i >>= 8;
	}

}

void pack_float(char *otp, char *str, char len)
{
	unsigned char sign = 0, mode = 0, wait, *os = otp;
	unsigned int n;
	unsigned _int64 m;
	double f, fm = 0;
	unsigned int j, k, sz, mn = mantissa[len], en = exponent[len];
	int e, em = 0, cm = 0;
	int len2 = tsz[len];

	if (*str == '-')
	{
		sign = 128;
		++str;
	}
	sz = strlen(str);

	for (j = 0; j < sz; ++j)
		if (str[j] == '.')
		{
			for (k = j+1; k < sz; ++k)
				if (str[k] == 'E' || str[k] == 'e')
				{
					f = atoi(&str[k+1]) / LOG_10_2;
					em = f;
					fm = fmod(f, 1);
					cm = n ^ (1 << em);
					str[k] = '\0';
					break;
				}
			f = atof(&str[j]);
			str[j] = '\0';
			n = atoi64(str);
			break;
		}
		else if (str[j] == 'E' || str[j] == 'e')
		{
			f = atoi(&str[j+1]) / LOG_10_2;
			em = f;
			fm = fmod(f, 1);
			cm = n ^ (1 << em);
			str[j] = '\0';
		}
	if (j == sz)
	{
		f = 0.0;
		n = atoi64(str);
	}

	if (!n && !f)
	{
		e = 0;
		m = 0;
	}
	else
	{

		fm = pow(2, fm);
		f = f * fm + fmod(n * fm, 1);
		n *= fm;
		if (f > 1)
		{
			--f;
			++n;
		}
		if (n)
		{
			e = log2(n);
			m = n ^ (1 << e);
			sz = mn - e - 1;
		}
		else
		{
			e = -1;
			m = 0;
			sz = mn;
		}
		e += em;

		wait = !n;
		m <<= 1;
		if (!f)
			m <<= sz;
		else for (j = 0; j < sz; ++j)
		{
			f *= 2;
			if (f >= 1)
			{
				if (!wait)
					++m;
				--f;
				wait = 0;
			}
			else if (wait)
			{
				--e;
				++sz;
			}
			m <<= 1;
		}
		if (f * 4 >= 1)
			++m;

		e += bias[len];
	}

	for (j = 0; j < len2; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				*otp++ = m & 255;
				m >>= 8;
				mn -= 8;
			}
			else
			{
				*otp++ = ((e & (255 >> mn)) << mn) | m;
				en -= 8 - mn;
				e >>= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en >= 8)
			{
				*otp++ = e & 255;
				e >>= 8;
				en -= 8;
			}
			else
			{
				*otp++ = sign | e;
			}
			break;
		}
	}
}

void print_err(file_info *fi, int pos, char *str)
{
	int j;

	errno = 1;
	if ((!asm_code && !inc_level) || muzzle)
		return;
	for (j = 0; j < fi->n_lines; ++j)
		if (pos < fi->lines[j].pos)
			break;
	--j;

	printf("%s:%d: %s\n", ((file_info *)(fi->lines[j].file))->name,
		fi->lines[j].line, str);
	++errcount;
}

void print_warn(file_info *fi, int pos, char *str)
{
	int j;

	if (muzzle)
		return;

	for (j = 0; j < fi->n_lines; ++j)
		if (pos < fi->lines[j].pos)
			break;
	--j;

	printf("%s:%d: Warning: %s\n", ((file_info *)(fi->lines[j].file))->name,
		fi->lines[j].line, str);
	++warncount;
}

void write_asm(char *str);

void print_var_info(var *vp)
{
	char buf[WORD_LENGTH];
	int j;

	sprintf(buf, ";\tVar: %s\n;\t\tPtr: %d", vp->name, vp->ptr);
	write_asm(buf);
	sprintf(buf, ";\t\tLoc: %d", vp->loc);
	write_asm(buf);
	for (j = 0; j < N_ATTRS; ++j)
		if (vp->attrs & (1 << j))
		{
			sprintf(buf, ";\t\tAttribute: %s", attrs[j]);
			write_asm(buf);
		}
}

void print_var_info_stdout(var *vp)
{
	int j;

	printf("\tVar: %s\n\t\tPtr: %d\n", vp->name, vp->ptr);
	printf("\t\tLoc: %d\n", vp->loc);
	for (j = 0; j < N_ATTRS; ++j)
		if (vp->attrs & (1 << j))
			printf("\t\tAttribute: %s\n", attrs[j]);
}

char *new_label()
{
	char buf[WORD_LENGTH];
	lstr *ls = (lstr *)calloc(1, sizeof(lstr));

	ls->next = label_base;
	if (label_base)
		label_base->prev = ls;
	label_base = ls;

	sprintf(buf, "L%d", ++n_labels);
	ls->str = (char *)calloc(strlen(buf)+1, sizeof(char));
	strcpy(ls->str, buf);

	return ls->str;
}

int open_file(file_info *fi, char *fpath, char *name)
{
	FILE *fp;
	struct stat st;
	int ch, pos = 0;

	fi->name = NULL;
	fi->text = NULL;
	fi->lines = NULL;
	fi->n_lines = 0;

	if (stat(fpath, &st) == -1)
		return 1;
	if (!(fp = fopen(fpath, "r")))
		return 1;

	fi->size = st.st_size;
	fi->name = (char *)calloc(strlen(name)+1, sizeof(char));
	strcpy(fi->name, name);
	fi->text = (char *)calloc(st.st_size+1, sizeof(char));
	fi->lines = (line_info *)calloc(1, sizeof(line_info));
	fi->n_lines = 1;
	
	while ((ch = fgetc(fp)) != EOF)
	{
		fi->text[pos++] = ch;
		if (ch == '\n')
		{
			fi->lines = (line_info *)realloc(fi->lines,
				++fi->n_lines * sizeof(line_info));
			fi->lines[fi->n_lines-1].line = fi->n_lines;
			fi->lines[fi->n_lines-1].pos = pos-1;
			fi->lines[fi->n_lines-1].file = fi;
		}
	}
	fi->text[pos++] = '\n';
	fi->text[pos] = '\0';

	fclose(fp);

	return 0;
}

file_info *fetch_file(char *fpath, char *name)
{
	file_info *fp;

	if (fpath)
	{
		standardize_path(fpath);
		for (fp = flist_base; fp; fp = fp->next)
		{
			if (!strcmp(fp->name, fpath))
				return fp;
		}
	}

	fp = (file_info *)calloc(1, sizeof(file_info));
	fp->next = flist_base;
	if (flist_base)
		flist_base->prev = fp;

	if (fpath)
	{
		if (open_file(fp, fpath, name))
		{
			free(fp);
			return NULL;
		}
	}
	else
	{
		fp->name = (char *)calloc(strlen(name)+1, sizeof(char));
		strcpy(fp->name, name);
		fp->lines = (line_info *)malloc(0);
	}

	return fp;
}

void mkstring(char *src, char *dest)
{
	*dest++ = '"';
	for (; *src; ++src)
	{
		switch (*src)
		{
		case '\\':
			*dest++ = '\\';
			*dest++ = '\\';
			break;
		case '"':
			*dest++ = '\\';
			*dest++ = '"';
			break;
		case '\n':
			*dest++ = '\\';
			*dest++ = 'n';
			break;
		case '\r':
			*dest++ = '\\';
			*dest++ = 'r';
			break;
		case '\t':
			*dest++ = '\\';
			*dest++ = 't';
			break;
		case '\b':
			*dest++ = '\\';
			*dest++ = 'b';
			break;
		default:
			*dest++ = *src;
		}
	}
	*dest++ = '"';
	*dest = '\0';
}

int macro_defined(char *str)
{
	macro *m;

	for (m = macro_base; m; m = m->next)
		if (!strcmp(str, m->name))
			return 1;

	return 0;
}

lstr *um_base;
int replace_macros(char *src, char *dest, char def, char *fname,
				   int line)
{
	int j, sz = strlen(src), k, k2, sk, paren, n, sv, had_args = 0;
	macro *m;
	char rep[WORD_LENGTH], *rp, flag, flag2, quote, squote, slash;
	char sword[WORD_LENGTH], sword2[WORD_LENGTH], *sp;
	char space, copy, pound, mode, swordy[WORD_LENGTH], flag3;
	lstr *ls;
	int ret = 0;

	for (j = 0; j <= sz; ++j)
	{
		sp = sword;
		for (; src[j] != '_' && !isalnum(src[j]); ++j)
		{
			if (j >= sz)
				goto RMD;
			*dest++ = src[j];
		}
		for (k2 = j; src[k2] == '_' || isalnum(src[k2]); ++k2)
			*sp++ = src[k2];
		*sp = '\0';
		flag = 1;

		if ((j ? isspace(src[j-1]) : 1) &&
			!strncmp(&src[j], "defined", 7) &&
			(j < sz-7 ? (isspace(src[j+7]) || src[j+7] == '(')
			: 0))
		{
			flag = 0;
			sp = sword;
			for (; j < sz; ++j)
				if (!isspace(src[j]))
					break;
			if (mode = (src[j] == '('))
				++j;
			for (; j < sz; ++j)
			{
				if ((mode && src[j] == ')') ||
					(!mode && isspace(src[j])))
					break;
				else
					*sp++ = src[j];
			}
			++j;
			*sp = '\0';
			trim_spaces(sword);
			*dest++ = '0' + macro_defined(sword);
			*dest++ = ' ';
		}
		else if (!strcmp(sword, "__LINE__"))
		{
			flag = 0;
			sprintf(sword, "%d", line);
			strcpy(dest, sword);
			dest += strlen(sword);
		}
		else if (!strcmp(sword, "__FILE__"))
		{
			flag = 0;
			strcpy(dest, fname);
			dest += strlen(fname);
		}
		else if (!strcmp(sword, "__DATE__"))
		{
			flag = 0;
			strcpy(dest, date_str);
			dest += strlen(date_str);
		}
		else if (!strcmp(sword, "__TIME__"))
		{
			flag = 0;
			strcpy(dest, time_str);
			dest += strlen(time_str);
		}
		else for (m = macro_base; m; m = m->next)
		{
			if (!strcmp(sword, m->name))
			{
				for (ls = um_base; ls; ls = ls->next)
					if (!strcmp(ls->str, m->name))
					{
						for (sp = m->name; *sp; ++sp)
							*dest++ = *sp;
						goto SubEnd;
					}
				flag = 0;

				ls = (lstr *)calloc(1, sizeof(lstr));
				ls->str = m->name;
				ls->next = um_base;
				um_base = ls;

				ret = 1;
				space = 1;
				for (k = j + strlen(m->name); k < sz; ++k)
				{
					if (src[k] == '(')
					{
						had_args = 1;
						ls = m->args;
						quote = squote = slash = paren = pound = 0;
						rp = rep;
						for (++k; k <= sz; ++k)
						{
							if (quote)
							{
								copy = 1;
								if (slash)
									slash = 0;
								else if (src[k] == '\\')
									slash = 1;
								else if (src[k] == '"')
									quote = 0;
							}
							else if (squote)
							{
								copy = 1;
								if (slash)
									slash = 0;
								else if (src[k] == '\\')
									slash = 1;
								else if (src[k] == '\'')
									squote = 0;
							}
							else if (src[k] == '"')
							{
								copy = 1;
								quote = 1;
							}
							else if (src[k] == '\'')
							{
								copy = 1;
								squote = 1;
							}
							else if (paren)
							{
								copy = 1;
								if (src[k] == '(' || src[k] == '[')
									++paren;
								else if (src[k] == ')' || src[k] == ']')
									--paren;
							}
							else if (src[k] == '(' || src[k] == '[')
							{
								copy = 1;
								paren = 1;
							}
							else if (src[k] == ',' || src[k] == ')')
							{
								copy = 0;
								if (!ls)
								{
									printf("Too many arguments to macro %s\n",
										m->name);
									return 1;
								}
								else
								{
									*rp = '\0';
									trim_spaces(rep);
									if (ls->rep)
										free(ls->rep);
									ls->rep = (char *)calloc(
										strlen(rep)+1, sizeof(char));
									strcpy(ls->rep, rep);
									rp = rep;
									ls = ls->next;
								}
								if (src[k] == ')')
								{
									if (ls)
									{
										printf("Not enough arguments to macro %s\n",
											m->name);
										return 1;
									}
									for (j = k; src[j]; ++j)
										if (src[j] == '_' ||
											isalnum(src[j]))
										{
											--j;
											break;
										}
									goto DoRep;
								}
							}
							else
								copy = 1;

							if (copy)
								*rp++ = src[k];
						}
					}
					else if (!isspace(src[k]))
						break;
				}

DoRep:			rp = rep;
				sk = strlen(m->value);
				pound = 0;
				sv = k+1;
				sp = swordy;
				for (k = 0; k <= sk; ++k)
				{
					flag2 = flag3 = 1;
					if (m->value[k] != '_' && !isalnum(m->value[k]) && swordy[0])
					{
						*sp = '\0';
						for (ls = m->args; ls; ls = ls->next)
						{
							if (!strcmp(swordy, ls->str))
							{
								switch (pound)
								{
								case 1:
									for (n = k; m->value[n] != '#'; --n)
										--rp;
									strcpy(sword2, ls->rep);
									trim_spaces(sword2);
									mkstring(sword2, sword);
									strcpy(rp, sword);
									rp += strlen(sword);
									break;
								case 0:
									strcpy(rp, ls->rep);
									rp += strlen(ls->rep);
									break;
								}
								flag3 = 0;
								pound = 0;
								break;
							}
						}

						if (flag3)
						{
							strcpy(rp, swordy);
							rp += strlen(swordy);
						}

						sp = swordy;
					}
					else if (m->value[k] == '_' || isalnum(m->value[k]))
					{
						flag2 = 0;
						*sp++ = m->value[k];
					}

					if (pound == 2)
					{
						if (isspace(m->value[k]))
						{
							flag2 = 0;
							goto MakeRep;
						}
						else
							pound = 0;
					}
					else if (m->value[k] == '#')
					{
						if (++pound > 2)
							pound = 0;
						else if (pound == 2)
						{
							--rp;
							for (n = k-2; n >= 0 &&
								isspace(m->value[n]); --n)
								--rp;
							flag2 = 0;
							goto MakeRep;
						}
					}

MakeRep:			if (flag2)
						*rp++ = m->value[k];
				}
				if (had_args && sv < sz)
					strcat(rep, &src[sv]);

				while (replace_macros(rep, sword, def, fname, line))
					strcpy(rep, sword);
				strcpy(dest, sword);
				dest += strlen(sword);
SubEnd:			flag = 0;
				break;
			}
		}
		if (had_args);
		else if (flag)
		{
			for (; j <= k2; ++j)
				*dest++ = src[j];
			--j;
		}
		else
			j = k2-1;
	}
RMD:
	*dest = '\0';

	return ret;
}

int is_float(char *str)
{
	int j, s = strlen(str), point = (*str == '.');
	char e = 0, *sp = str;

	if (!s) return 0;
	if (*str != '-' && !isdigit(*str) && *str != '.') return 0;
	if ((*str == '-' || *str == '.') && !str[1]) return 0;
	for (j = 1; j < s; ++j)
		if (e)
		{
			if (!isdigit(str[j]) && !((str[j] == '+' || str[j] == '-')
				&& e == 1) && str[j] != 'L' && str[j] != 'l'
				&& str[j] != 'F' && str[j] != 'f')
				return 0;
			e = 2;
		}
		else
		{
		if (str[j] == '.')
			++point;
		else if (str[j] == 'E' || str[j] == 'e')
			e = 1;
		else if (!isdigit(str[j]) && str[j] != 'L' && str[j] != 'l'
			&& str[j] != 'F' && str[j] != 'f')
			return 0;
		}
	return point == 1 || (e && point <= 1);
}

int is_opchars(char *str)
{
	char *os, flag;

	for (; *str; ++str)
	{
		flag = 1;
		for (os = op_chars; *os; ++os)
			if (*str == *os)
			{
				flag = 0;
				break;
			}
		if (flag)
			return 0;
	}
	
	return 1;
}

int is_int(char *str)
{
	if (*str == '-' && !*(str+1))
		return 0;
	if (*str != '-' && !isdigit(*str))
		return 0;
	if (*str == '-')
	{
		if (str[1] == '0' && str[2])
			return 0;
	}
	else
	{
		if (*str == '0' && str[1])
			return 0;
	}

	for (++str; *str; ++str)
		if (!isdigit(*str) && *str != 'U' && *str != 'u' &&
			*str != 'L' && *str != 'l')
			return 0;
	return 1;
}

int is_hex(char *str)
{
	if (str[0] != '0' || str[1] != 'x' || !str[2])
		return 0;
	for (str += 2; *str; ++str)
		if (!isxdigit(*str) && *str != 'U' && *str != 'u' &&
			*str != 'L' && *str != 'l')
			return 0;
	return 1;
}

int is_octal(char *str)
{
	if (str[0] != '0' || !str[1])
		return 0;
	for (++str; *str; ++str)
		if ((*str < '0' || *str > '7')  && *str != 'U' && *str != 'u' &&
			*str != 'L' && *str != 'l')
			return 0;
	return 1;
}

int is_any_int(char *str)
{
	return is_int(str) || is_hex(str) || is_octal(str);
}

int htoi(char *str)
{
	int j, ret = 0, base = 1;

	for (j = strlen(str)-1; j > 1; --j)
	{
		if (str[j] >= 'A' && str[j] <= 'F')
			ret += (str[j] - 'A' + 10) * base;
		else if (str[j] >= 'a' && str[j] <= 'f')
			ret += (str[j] - 'a' + 10) * base;
		else
			ret += (str[j] - '0') * base;
		base *= 16;
	}

	return ret;
}

int otoi(char *str)
{
	int j, ret = 0, base = 1;

	for (j = strlen(str)-1; j; --j)
	{
		ret += (str[j] - '0') * base;
		base *= 8;
	}

	return ret;
}

int anytoi(char *str)
{
	if (is_hex(str))
		return htoi(str);
	else if (is_octal(str))
		return otoi(str);
	else
		return atoi(str);
}

int is_complete(char *str, char unary)
{
	char paren = 0, quote = 0, squote = 0, slash = 0, start = 1;

	for (; *str; ++str)
	{
		if (slash)
			slash = 0;
		else if (quote)
		{
			if (*str == '\\')
				slash = 1;
			else if (*str == '"')
				quote = 0;
		}
		else if (squote)
		{
			if (*str == '\\')
				slash = 1;
			else if (*str == '\'')
				squote = 0;
		}
		else if (paren)
		{
			if (*str == '(' || *str == '[')
				++paren;
			else if (*str == ')' || *str == ']')
				--paren;
		}
		else if (*str == '"')
		{
			quote = 1;
			start = 0;
		}
		else if (*str == '\'')
		{
			squote = 1;
			start = 0;
		}
		else if (*str == '(' || *str == '[')
		{
			paren = 1;
			start = 0;
		}
		else if (is_opchar(*str))
		{
			if (*str == '.' || (*str == '-' && *(str+1) == '>')
				|| (*str == '>' && *(str-1) == '-'))
				continue;
			if (unary && start)
				if (*str == '-' && *str == '+' || *str == '*' ||
					*str == '&' || *str == '~' || *str == '!')
					continue;
			return 0;
		}
	}

	return 1;
}

int is_cast_chars(char *str)
{
	for (; *str; ++str)
		if (*str != '*' && *str != '(' && *str != ')' &&
			*str != '[' && *str != ']' && !isspace(*str))
			return 0;
	return 1;
}

int is_cast(char *str, code *c)
{
	char sword[WORD_LENGTH], *sp = sword, paren = 1;
	char buf[WORD_LENGTH];
	int j;

	if (*str != '(')
		return 0;
	for (++str; *str; ++str)
	{
		if (*str == '(' || *str == '[')
		{
			++paren;
			*sp++ = *str;
		}
		else if (*str == ')' || *str == ']')
		{
			if (!--paren)
			{
				*sp = '\0';
				break;
			}
		}
		else
			*sp++ = *str;
	}

	if (!is_complete(++str, 1))
		return 0;

	first_word(buf, sword);
	if (find_td(buf, c))
		return 1;
	for (j = 0; j < N_ATTRS; ++j)
		if (!strcmp(buf, attrs[j]))
			return 1;
	return 0;
}

int is_arrexp(char *str)
{
	if (!*str || !is_complete(str, 0))
		return 0;

	if (str[strlen(str)-1] != ']')
		return 0;

	return 1;
}

int is_char(char *str)
{
	char slash = 0;
	int j, sz = strlen(str);

	if (sz < 3)
		return 0;
	if (str[0] != '\'' || str[sz-1] != '\'')
		return 0;
	--sz;

	for (j = 1; j < sz; ++j)
	{
		if (slash)
			slash = 0;
		else if (str[j] == '\\')
			slash = 1;
		else if (str[j] == '\'')
			return 0;
	}

	return 1;
}

int translate_char(char *str, char *ch)
{
	int n;

	if (*str == '\\')
	{
		switch (*(str+1))
		{
		case 'n':
			*ch = '\n';
			return 2;
		case 'r':
			*ch = '\r';
			return 2;
		case 'b':
			*ch = '\b';
			return 2;
		case 't':
			*ch = '\t';
			return 2;
		case 'v':
			*ch = '\v';
			return 2;
		case 'a':
			*ch = '\v';
			return 2;
		case 'f':
			*ch = '\f';
			return 2;
		case 'x':
			*ch = 0;
			n = 2;
			for (str += 2; isxdigit(*str); ++str)
			{
				++n;
				*ch <<= 4;
				if (isdigit(*str))
					*ch += *str - '0';
				else if (isupper(*str))
					*ch += *str - 'A';
				else
					*ch += *str - 'a';
			}
			return n;
		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
			*ch = 0;
			n = 1;
			for (++str; *str >= '0' && *str <= '7'; ++str)
			{
				++n;
				*ch <<= 3;
				*ch += *str - '0';
			}
			return n;
		default:
			*ch = *(str+1);
			return 2;
		}
	}
	else
	{
		*ch = *str;
		return 1;
	}
}

int is_string(char *str)
{
	char slash = 0;
	int j, sz = strlen(str);

	if (sz < 2)
		return 0;
	if (str[0] != '"' || str[sz-1] != '"')
		return 0;
	--sz;

	for (j = 1; j < sz; ++j)
	{
		if (slash)
			slash = 0;
		else if (str[j] == '\\')
			slash = 1;
		else if (str[j] == '"')
			return 0;
	}

	return 1;
}

int translate_string(char *buf, char *str)
{
	char slash = 0;
	int ret = 1, n;

	for (; *str; ++str)
	{
		n = translate_char(str, buf++);
		++ret;
		str += n - 1;
	}

	*buf = '\0';
	return ret;
}

int is_fcall(char *str)
{
	int paren = 1, brack = 0;
	char quote = 0, squote = 0, slash = 0, *sp = str;

	if (!isalnum(*str) && *str != '_')
		return 0;
	for (++str; *str; ++str)
	{
		if (brack)
		{
			if (*str == '(' || *str == '[')
				++brack;
			else if (*str == ')' || *str == ']')
				--brack;
		}
		else if (*str == '[')
			++brack;
		else if (!isalnum(*str) && *str != '_' && *str != '.' &&
			!(*str == '-' && *(str+1) == '>') && !(*(str-1) == '-'
			&& *str == '>'))
			break;
	}
	for (; *str == ' '; ++str);

	if (*str != '(' || str[strlen(str)-1] != ')')
		return 0;

	for (++str; *str; ++str)
	{
		if (!paren)
			return 0;
		else if (quote)
		{
			if (slash)
				slash = 0;
			else if (*str == '\\')
				slash = 1;
			else if (*str == '"')
				quote = 0;
		}
		else if (squote)
		{
			if (slash)
				slash = 0;
			else if (*str == '\\')
				slash = 1;
			else if (*str == '\'')
				squote = 0;
		}
		else if (*str == '"')
			quote = 1;
		else if (*str == '\'')
			squote = 1;
		else if (*str == '[')
			++brack;
		else if (*str == ']')
		{
			if (--brack < 0)
				return 0;
		}
		else if (*str == '(')
			++paren;
		else if (*str == ')')
			--paren;
	}

	return 1;
}

int is_sizeof(char *str)
{
	int paren = 0, stuff = 0;

	if (strncmp(str, "sizeof", 6))
		return 0;
	for (str += 6; *str; ++str)
	{
		if (paren)
		{
			if (*str == '(' || *str == '[')
				++paren;
			else if (*str == ')' || *str == ']')
			{
				stuff = 2;
				--paren;
			}
		}
		else if (stuff == 1)
		{
			if (*str == ' ')
				stuff = 2;
		}
		else if (stuff == 2)
		{
			if (*str != ' ')
				return 0;
		}
		else if (*str == '(')
			stuff = paren = 1;
		else if (*str != ' ')
			stuff = 1;
	}

	return 1;
}

void del_asm()
{
	if (asm_code->cur_asm == asm_code->asm_base)
	{
		free(asm_code->cur_asm->str);
		free(asm_code->cur_asm);
		asm_code->asm_base = asm_code->cur_asm = NULL;
	}
	else
	{
		asm_code->cur_asm = asm_code->cur_asm->prev;
		asm_code->cur_asm->next = NULL;
	}
}

void write_asm(char *str)
{
	asmt *a;

	if (muzzle || !asm_code)
		return;

	if (asm_code->cur_asm)
	{
		if (!strncmp(asm_code->cur_asm->str, "push", 4) && !strncmp(str, "pop", 3))
		{
			if (asm_code->cur_asm->str[strlen(asm_code->cur_asm->str)-1] ==
				str[strlen(str)-1])
			{
				del_asm();
			}
			else
			{
				if (asm_code->cur_asm->prev)
				{
					if (!strncmp(asm_code->cur_asm->prev->str,
						"ld", 2))
					{
						if (asm_code->cur_asm->prev->str[2] == ' ')
						{
							if (asm_code->cur_asm->prev->str[3] == asm_code->cur_asm->str[strlen(asm_code->cur_asm->str)-1])
							{
								asm_code->cur_asm->prev->str[3] = str[strlen(str)-1];
								del_asm();
								return;
							}
						}
						else if (asm_code->cur_asm->prev->str[4] == asm_code->cur_asm->str[strlen(asm_code->cur_asm->str)-1])
						{
							asm_code->cur_asm->prev->str[4] = str[strlen(str)-1];
							del_asm();
							return;
						}
					}
				}
				sprintf(asm_code->cur_asm->str, "ld8 %c,%c", str[strlen(str)-1],
					asm_code->cur_asm->str[strlen(asm_code->cur_asm->str)-1]);
			}
			return;
		}
	}

	a = (asmt *)calloc(1, sizeof(asmt));
	a->str = (char *)calloc(strlen(str)+1, sizeof(char));
	strcpy(a->str, str);

	if (asm_code->asm_base)
	{
		a->prev = asm_code->cur_asm;
		asm_code->cur_asm->next = a;
	}
	else
		asm_code->asm_base = a;
	asm_code->cur_asm = a;
}

void queue_asm(code *c, char *str)
{
	asmt *a = (asmt *)calloc(1, sizeof(asmt));

	a->str = (char *)calloc(strlen(str)+1, sizeof(char));
	strcpy(a->str, str);

	if (c->aq_base)
	{
		a->prev = c->aq;
		c->aq->next = a;
	}
	else
		c->aq_base = a;
	c->aq = a;
}

#define PRINT_ERR(X) print_err(cur_file, cur_line, (X))
#define PRINT_WARN(X) print_warn(cur_file, cur_line, (X))

void set_sizes(var*);

int get_sn(int n)
{
	switch (n)
	{
	case 1:
		return 0;
	case 2:
		return 1;
	case 4:
		return 2;
	case 8:
		return 3;
	}
}

int roundoff(float n)
{
	if (n <= 1)
		return 1;
	else if (n <= 2)
		return 2;
	else if (n <= 4)
		return 4;
	else
		return 8;
}

int choose_cast(var *v1, var *v2)
{
	if (v1->ptr)
		return 1;
	else if (v2->ptr)
		return 2;
	else if (v1->attrs & (FLOAT | DOUBLE))
	{
		if (v2->attrs & (FLOAT | DOUBLE))
		{
			if (v1->size > v2->size)
				return 2;
			else
				return 1;
		}
		else
			return 2;
	}
	else if (v2->attrs & (FLOAT | DOUBLE))
		return 1;
	else if (v1->size < v2->size)
		return 1;
	else
		return 2;
}

int cast(var *src, var *targ)
{
	char buf[WORD_LENGTH];
	int i1, t1, t2;

	if (src->ptr || targ->ptr)
	{
		memcpy(src, targ, sizeof(var));
		return 1;
	}
	else if (src->fptr && targ->fptr)
	{
		if ((src->attrs & HOST) != (targ->attrs & HOST))
			PRINT_WARN("Casting between standard and host function pointers");
		else if (typecmp(src, targ))
			PRINT_WARN("Casting between incompatible function pointer types");
		memcpy(src, targ, sizeof(var));
		return 1;
	}

	if (src->attrs & (FLOAT | DOUBLE))
		t1 = _C_FLOAT;
	else if (src->attrs & SIGNED)
		t1 = _C_SIGNED;
	else
		t1 = _C_UNSIGNED;
	if (targ->attrs & (FLOAT | DOUBLE))
		t2 = _C_FLOAT;
	else if (targ->attrs & SIGNED)
		t2 = _C_SIGNED;
	else
		t2 = _C_UNSIGNED;

	i1 = get_sn(targ->unit_size);
	i1 = (i1 << 2) | t2;
	i1 <<= 2;
	i1 |= get_sn(src->unit_size);
	i1 = (i1 << 2) | t1;
	sprintf(buf, "cast %d", i1);
	write_asm(buf);

	memcpy(src, targ, sizeof(var));
	return 0;
}

_int64 eval(char *str, var *v, char pp, char lvalue)
{
	char arg1[WORD_LENGTH], arg2[WORD_LENGTH], opr[WORD_LENGTH];
	char earg[WORD_LENGTH], op_base[2] = "\0\0";
	char *ap1, *ap2, *op, sword[WORD_LENGTH], buf[WORD_LENGTH];
	char buf2[WORD_LENGTH];
	char status;
	int j, sz, k, n, i, i2, si, on, ons;
	_int64 a1, a2;
	var varg1, varg2, varg3, *vp, *arg;
	char quote, squote, slash, flag, *sp, *sp2, *sw, omuzzle, *lb1;
	char *lb2;
	int paren, tflag;
	lstr *ls, *ls_base;
	func *f;
	code *acbak;
	enumt *e;
	econst *ec;
	_float64 f1, f2;
	memset(v, 0, sizeof(var));

	v->bf_size = -1;
	trim_spaces(str);
	if (!(sz = strlen(str)))
	{
		write_asm("push a");
		return 0;
	}

	if (sz >= 2 && str[0] == '(' && str[sz-1] == ')')
	{
		paren = flag = 1;
		quote = squote = slash = 0;
		for (j = 1; j < sz-1; ++j)
		{
			if (quote)
			{
				if (slash)
					slash = 0;
				else if (str[j] == '\\')
					slash = 1;
				else if (str[j] == '"')
					quote = 0;
			}
			else if (squote)
			{
				if (slash)
					slash = 0;
				else if (str[j] == '\\')
					slash = 1;
				else if (str[j] == '\'')
					squote = 0;
			}
			else if (str[j] == '"')
				quote = 1;
			else if (str[j] == '\'')
				squote = 1;
			else if (str[j] == '(' || str[j] == '[')
				++paren;
			else if (str[j] == ')' || str[j] == ']')
				if (!--paren)
				{
					flag = 0;
					break;
				}
		}

		if (flag)
		{
			str[sz-1] = '\0';
			return eval(++str, v, pp, lvalue);
		}
	}

	sprintf(buf, "; Eval[%d]: %s", lvalue, str);
	write_asm(buf);
//	printf("Eval: %s\n", str);
	if (!pp && is_const(str, asm_code))
	{
		sprintf(buf, "; Constant: %s", str);
		write_asm(buf);
		a1 = eval(str, &varg1, 1, 0);
		if (varg1.attrs & (FLOAT | DOUBLE))
		{
			f1 = *(_float64 *)&a1;
			v->attrs = CONST | DOUBLE;
			set_sizes(v);
			if (fmod(f1, 1))
				sprintf(buf, "ld8 a,%g", f1);
			else
				sprintf(buf, "ld8 a,%g.0", f1);
			write_asm(buf);
			write_asm("push8 a");
		}
		else
		{
			v->attrs = CONST | SIGNED | INT;
			set_sizes(v);
			sprintf(buf, "ld a,%d", a1);
			write_asm(buf);
			write_asm("push a");
		}
		return 0;
	}
	else if (is_int(str))
	{
		v->attrs = SIGNED | INT | CONST;
		for (j = strlen(str)-1; j >= 0; --j)
			if (str[j] == 'L' || str[j] == 'l')
			{
				v->attrs |= LONG;
				str[j] = '\0';
			}
			else if (str[j] == 'U' || str[j] == 'u')
			{
				if (v->attrs & SIGNED)
					v->attrs ^= SIGNED;
				v->attrs |= UNSIGNED;
				str[j] = '\0';
			}
			else
				break;
		set_sizes(v);

		if (pp)
			return atoi(str);
		else
		{
			sprintf(buf, "ld%d a,%s", v->load_size, str);
			write_asm(buf);
			sprintf(buf, "push%d a", v->load_size);
			write_asm(buf);

			return 0;
		}
	}
	else if (is_hex(str))
	{
		v->attrs = SIGNED | INT | CONST;
		for (j = strlen(str)-1; j >= 0; --j)
			if (str[j] == 'L' || str[j] == 'l')
			{
				v->attrs |= LONG;
				str[j] = '\0';
			}
			else if (str[j] == 'U' || str[j] == 'u')
			{
				if (v->attrs & SIGNED)
					v->attrs ^= SIGNED;
				v->attrs |= UNSIGNED;
				str[j] = '\0';
			}
			else
				break;
		set_sizes(v);

		n = htoi(str);
		if (pp)
			return n;
		else
		{
			sprintf(buf, "ld%d a,%u", v->load_size, n);
			write_asm(buf);
			sprintf(buf, "push%d a", v->load_size);
			write_asm(buf);

			return 0;
		}
	}
	else if (is_octal(str))
	{
		v->attrs = SIGNED | INT | CONST;
		for (j = strlen(str)-1; j >= 0; --j)
			if (str[j] == 'L' || str[j] == 'l')
			{
				v->attrs |= LONG;
				str[j] = '\0';
			}
			else if (str[j] == 'U' || str[j] == 'u')
			{
				if (v->attrs & SIGNED)
					v->attrs ^= SIGNED;
				v->attrs |= UNSIGNED;
				str[j] = '\0';
			}
			else
				break;
		set_sizes(v);

		n = otoi(str);
		if (pp)
			return n;
		else
		{
			sprintf(buf, "ld%d a,%u", v->load_size, n);
			write_asm(buf);
			sprintf(buf, "push%d a", v->load_size);
			write_asm(buf);

			return 0;
		}
	}
	else if (is_float(str))
	{
		v->attrs = CONST | DOUBLE;

		for (j = strlen(str)-1; j >= 0; --j)
			if (str[j] == 'F' || str[j] == 'f')
			{
				v->attrs = CONST | FLOAT;
				str[j] = '\0';
			}
			else if (str[j] == 'L' || str[j] == 'l')
			{
				v->attrs = CONST | LONG | DOUBLE;
				str[j] = '\0';
			}
			else
				break;

		if (pp)
		{
			f1 = atof(str);
			return *(_int64 *)&f1;
		}

		set_sizes(v);
		sprintf(buf, "ld%d a,%s", v->load_size, str);
		write_asm(buf);
		sprintf(buf, "push%d a", v->load_size);
		write_asm(buf);
		return 0;
	}
	else if (is_char(str))
	{
		i = translate_char(++str, &flag);
		if (i != strlen(str)-1)
		{
			PRINT_ERR("Invalid `char' constant");
			printf("%d, %d\n", i, strlen(str)-1);
			return 1;
		}
		if (pp)
			return flag;
		sprintf(buf, "ld1 a,%d", flag);
		write_asm(buf);
		write_asm("push1 a");

		v->attrs = SIGNED | CHAR | CONST;
		set_sizes(v);
		return 0;
	}
	else if (is_string(str))
	{
		str[strlen(str)-1] = '\0';
		n = translate_string(sword, str+1);

		sprintf(buf2, ".db %u", *(sp = sword));
		if (sp)
		{
			for (++sp; *sp; ++sp)
			{
				sprintf(buf, ",%u", *sp);
				strcat(buf2, buf);
			}
			strcat(buf2, ",0");
		}

		sprintf(buf, "ld a,%d", global.code_base->self_size);
		write_asm(buf);
		write_asm("push a");
			
		acbak = asm_code;
		asm_code = global.code_base;
		write_asm(buf2);
		sprintf(buf, "ld a,%d", asm_code->self_size);
		write_asm(buf);
		sprintf(buf, "ld b,%d", cur_data++);
		write_asm(buf);
		write_asm("dcpy a,b");
		asm_code->self_size += n;
		asm_code = acbak;

		v->attrs = CONST | SIGNED | CHAR;
		v->ptr = 1;
		set_sizes(v);
		return 0;
	}
	else if (is_sizeof(str))
	{
		str += 6;
		trim_spaces(str);
		if (*str == '(')
		{
			str[strlen(str)-1] = '\0';
			++str;
		}
		trim_spaces(str);

		first_word(buf, str);
		if (find_td(buf, asm_code))
			goto szt;
		for (j = 0; j < N_ATTRS; ++j)
			if (!strcmp(buf, attrs[j]))
			{
szt:			parse_declr(str, &varg1, NULL, NULL, 0);
				if (pp)
					return varg1.size;
				sprintf(buf, "ld a,%d", varg1.size);
				write_asm(buf);
				write_asm("push a");
				v->attrs = CONST | SIGNED | INT;
				set_sizes(v);
				return 0;
			}

		omuzzle = muzzle;
		muzzle = 1;
		eval(str, &varg1, 0, 0);
		muzzle = omuzzle;
		set_sizes(&varg1);

		if (pp)
			return varg1.size;

		sprintf(buf, "ld a,%d", varg1.size);
		write_asm(buf);
		write_asm("push a");
		v->attrs = CONST | SIGNED | INT;
		set_sizes(v);

		return 0;
	}
	else if (is_cast(str, asm_code))
	{
		paren = 1;
		for (sp = ++str; *sp; ++sp)
		{
			if (*sp == '(')
				++paren;
			else if (*sp == ')')
			{
				if (!--paren)
				{
					*sp++ = '\0';
					break;
				}
			}
		}

		eval(sp, &varg1, 0, 0);
		parse_declr(str, &varg2, NULL, NULL, 0);
		sprintf(buf, "; %d[%s]->%d[%s]", varg1.unit_size,
			sp, varg2.unit_size, str);
		write_asm(buf);
		cast(&varg1, &varg2);
		memcpy(v, &varg1, sizeof(var));
		return 0;
	}
	else if (is_arrexp(str))
	{
		str[strlen(str)-1] = '\0';
		sp = &str[strlen(str)-1];
		paren = 1;
		for (; sp != str; --sp)
		{
			if (*sp == ']' || *sp == ')')
				++paren;
			else if (*sp == '[' || *sp == '(')
				if (!--paren)
					break;
		}
		*sp = '\0';
		++sp;
		eval(str, &varg1, 0, 1);
		eval(sp, &varg2, 0, 0);

		if (!varg1.arr_base && !varg1.ptr)
		{
			PRINT_ERR("Non-array/pointer expression used with array subscript");
			return 1;
		}

		if (varg1.arr_base)
			varg1.arr_base = varg1.arr_base->next;
		--varg1.ptr;
		set_sizes(&varg1);

		write_asm("pop a");
		if (varg1.size > 1)
		{
			sprintf(buf, "ld b,%d", varg1.size);
			write_asm(buf);
			write_asm("mult a,b");
		}
		write_asm("pop b");
		write_asm("add a,b");

		if (lvalue || varg1.arr_base)
		{
			write_asm("push a");
			++varg1.ptr;
			set_sizes(&varg1);
		}
		else
		{
			sprintf(buf, "ld%d b,(a)", varg1.load_size);
			write_asm(buf);
			sprintf(buf, "push%d b", varg1.load_size);
			write_asm(buf);
		}

		memcpy(v, &varg1, sizeof(var));
		return 0;
	}
	else if (is_valid_id(str))
	{
		if (vp = find_var(str, asm_code))
		{
			if ((vp->attrs & (STRUCT | UNION)) && !(lvalue || vp->ptr || vp->arr_base))
			{
				PRINT_ERR("`struct' or `union' not permitted as rvalue expression");
				return 1;
			}

			memcpy(v, vp, sizeof(var));
			if (v->global)
			{
				if (lvalue || v->arr_base)
				{
					sprintf(buf, "ld a,%d", v->loc);
					write_asm(buf);
					write_asm("push a");
					++v->ptr;
				}
				else
				{
					sprintf(buf, "ld%d a,(%d)", v->load_size,
						v->loc);
					write_asm(buf);
					sprintf(buf, "push%d a", v->load_size);
					write_asm(buf);
				}
			}
			else
			{
				if (lvalue || (v->arr_base && !v->argarr))
				{
					sprintf(buf, "laddr %d", v->loc);
					write_asm(buf);
					++v->ptr;
				}
				else
				{
					sprintf(buf, "laddr %d", v->loc);
					write_asm(buf);
					write_asm("pop b");
					sprintf(buf, "ld%d a,(b)", v->load_size);
					write_asm(buf);
					sprintf(buf, "push%d a", v->load_size);
					write_asm(buf);
				}
			}

			set_sizes(v);
			return 0;
		}
		else if (ec = find_econst(str, asm_code, &e))
		{
			if (pp)
				return ec->val;

			sprintf(buf, "ld a,%d", ec->val);
			write_asm(buf);
			write_asm("push a");

			v->attrs = CONST | ENUM;
			v->t.e = e;
			set_sizes(v);
			return 0;
		}
		else if (f = find_func(str))
		{
			if (pp)
				return 0;

			if (f->type.attrs & HOST)
				sprintf(buf, "ld a,%d", f->type.host);
			else
				sprintf(buf, "ld a,_%s", str);
			write_asm(buf);
			write_asm("push a");

			memcpy(v, &f->type, sizeof(var));
			v->func = 0;
			v->fptr = 1;
			set_sizes(v);
			return 0;
		}
		else
		{
			sprintf(buf, "Parse error at `%s'", str);
			PRINT_ERR(buf);
			return 0;
		}
	}
	else if (is_fcall(str))
	{
		if (pp)
			return 0;

		str[strlen(str)-1] = ',';
		for (sp = str; *sp != '('; ++sp);
		*sp++ = '\0';

		if (!strcmp(str, "__va_arg__"))
		{
			sp[strlen(sp)-1] = '\0';
			parse_declr(sp, v, NULL, NULL, 0);
			return 0;
		}

		if (!(f = find_func(str)))
		{
			strcpy(buf, str);
			omuzzle = muzzle;
			muzzle = 1;
			eval(buf, &varg1, 0, 0);
			muzzle = omuzzle;
			if (!varg1.fptr)
			{
				PRINT_ERR("Attempt to call non-function expression");
				return 1;
			}
		}

		i = quote = squote = slash = paren = 0;
		sw = sword;
		ls_base = NULL;

		if (*sp == ',' && !*(sp+1));
		else for (; *sp; ++sp)
		{
			if (quote)
			{
				if (slash)
					slash = 0;
				else if (*sp == '\\')
					slash = 1;
				else if (*sp == '"')
					quote = 0;
				*sw++ = *sp;
			}
			else if (squote)
			{
				if (slash)
					slash = 0;
				else if (*sp == '\\')
					slash = 1;
				else if (*sp == '\'')
					squote = 0;
				*sw++ = *sp;
			}
			else if (paren)
			{
				if (*sp == '(' || *sp == '[')
					++paren;
				else if (*sp == ')' || *sp == ']')
					--paren;
				*sw++ = *sp;
			}
			else if (*sp == '"')
			{
				quote = 1;
				*sw++ = '"';
			}
			else if (*sp == '\'')
			{
				squote = 1;
				*sw++ = '\'';
			}
			else if (*sp == '(' || *sp == '[')
			{
				paren = 1;
				*sw++ = *sp;
			}
			else if (*sp == ',')
			{
				*sw = '\0';

				++i;
				ls = (lstr *)calloc(1, sizeof(lstr));
				ls->next = ls_base;
				if (ls_base)
					ls_base->prev = ls;
				ls_base = ls;
				ls->str = (char *)calloc(strlen(sword)+1,
					sizeof(char));
				strcpy(ls->str, sword);

				sw = sword;
			}
			else
				*sw++ = *sp;
		}

		vp = f ? &f->type : &varg1;
		for (arg = vp->args; arg; arg = arg->next_arg)
			if (!arg->next_arg)
				break;

		ls = ls_base;
		if (vp->vararg)
		{
			if (i < vp->nargs)
			{
				sprintf(buf, "Not enough arguments passed to function `%s'",
					str);
				PRINT_ERR(buf);
				return 1;
			}
			else if (i > vp->nargs)
			{
				if (!(vp->attrs & HOST))
				{
					i -= vp->nargs;
					for (j = 0; j < i; ++j)
					{
						eval(ls->str, &varg2, 0, 0);

						if (vp->attrs & HOST)
						{
							sprintf(buf, "pop%d a", varg2.load_size);
							write_asm(buf);
							sprintf(buf, "arg%d a", varg2.load_size);
							write_asm(buf);
						}

						ls = ls->next;
					}
				}
			}
		}
		else
		{
			if (i < vp->nargs)
			{
				sprintf(buf, "Not enough arguments passed to function `%s'",
					str);
				PRINT_ERR(buf);
				return 1;
			}
			else if (i > vp->nargs)
			{
				sprintf(buf, "Too many arguments passed to function `%s'",
					str);
				PRINT_ERR(buf);
				return 1;
			}
		}

		if (!(vp->attrs & HOST))
		{
			for (; ls; ls = ls->next)
			{
				eval(ls->str, &varg2, 0, 0);
				set_sizes(&varg2);
				ls->size = varg2.load_size;
				if (typecmp(&varg2, arg))
					cast(&varg2, arg);

				arg = arg->prev_arg;
			}
		}
		else
		{
			for (ls = ls_base; ls; ls = ls->next)
			{
				if (!is_const(ls->str, asm_code))
				{
					eval(ls->str, &varg2, 0, 0);
					if (arg)
						if (typecmp(&varg2, arg))
							cast(&varg2, arg);
					set_sizes(&varg2);
					ls->size = varg2.load_size;
				}
				else
					ls->flag = 1;

				if (arg)
					arg = arg->prev_arg;
			}

			for (ls = ls_base; ls; ls = ls->next)
				if (!ls->next)
					break;
			arg = vp->args;
//			for (ls = ls_base; ls; ls = ls->next)
			for (; ls; ls = ls->prev)
			{
				sprintf(buf, "; %d : %s", ls->flag, ls->str);
				write_asm(buf);
				if (!ls->flag)
				{
					sprintf(buf, "pop%d a", arg ? arg->load_size : ls->size);
					write_asm(buf);
				}
				if (settings.nonet);
				else if (vp->attrs & (INPUT | OUTPUT))
				{
					if (!ls->flag)
					{
						if (arg->attrs & STRING)
							write_asm("slen b,a");
						else if (arg->attrs & SYNC)
						{
							if (!arg->arr_base && arg->ptr)
							{
								memcpy(&varg2, arg, sizeof(var));
								--varg2.ptr;
								set_sizes(&varg2);
							}
							else
								varg2.size = arg->size;
							sprintf(buf, "ld b,%d", varg2.size);
							write_asm(buf);
						}
						write_asm("rmode");
						write_asm("rsend a");
						if (arg->attrs & (STRING | SYNC))
						{
							write_asm("rsend b");
							write_asm("msend a,b");
						}
						write_asm("emode");
					}
					flag = 1;
					if (arg->prev_arg)
					{
						if (!(arg->prev_arg->attrs & (SYNC | STRING)) && ls->flag)
							flag = 0;
					}
					if (flag)
					{
						if (vp->attrs & OUTPUT)
							write_asm("omode");
						else
							write_asm("imode");
					}
				}
				if (ls->flag)
				{
					a1 = eval(ls->str, &varg1, 1, 0);
					if (varg1.attrs & (FLOAT | DOUBLE))
					{
						f1 = *(_float64 *)&a1;
						if (!arg)
						{
							if (fmod(f1, 1))
								sprintf(buf, "arg%d %g",
									varg1.load_size, f1);
							else
								sprintf(buf, "arg%d %g.0",
									varg1.load_size, f1);
							write_asm(buf);
						}
						else if (arg->attrs & (FLOAT | DOUBLE))
						{
							if (fmod(f1, 1))
								sprintf(buf, "arg%d %g",
									arg->load_size, f1);
							else
								sprintf(buf, "arg%d %g.0",
									arg->load_size, f1);
							write_asm(buf);
						}
						else
						{
							sprintf(buf, "arg%d %d",
								arg->load_size, (int)f1);
							write_asm(buf);
						}
					}
					else
					{
						if (!arg)
						{
							if (varg1.attrs & (FLOAT | DOUBLE))
							{
								sprintf(buf, "arg%d %d.0",
									varg1.load_size, a1);
								write_asm(buf);
							}
							else
							{
								sprintf(buf, "arg%d %d ; Yes!",
									varg1.load_size, a1);
								write_asm(buf);
							}
						}
						else if (arg->attrs & (FLOAT | DOUBLE))
						{
							sprintf(buf, "arg%d %d.0",
								arg->load_size, a1);
							write_asm(buf);
						}
						else
						{
							sprintf(buf, "arg%d %d",
								arg->load_size, a1);
							write_asm(buf);
						}
					}
				}
				else
				{
					sprintf(buf, "arg%d a ; 1", arg ? arg->load_size : ls->size);
					write_asm(buf);
				}
				if (!settings.nonet && (vp->attrs & (OUTPUT | INPUT)))
				{
					flag = 1;
					if (arg->next_arg)
					{
						if (!(arg->next_arg->attrs & (SYNC | STRING)) && ls->prev->flag)
							flag = 0;
					}
					else
						flag = 0;
					if (flag)
						write_asm("emode");
				}
				if (arg)
					arg = arg->next_arg;
			}
		}

		free_lstr(ls_base);

		if (f)
		{
			if (f->type.attrs & HOST)
			{
				if (settings.nonet);
				else if (!f->type.args)
				{
					if (vp->attrs & INPUT)
						write_asm("imode");
					else if (vp->attrs & OUTPUT)
						write_asm("omode");
				}
				sprintf(buf, "hcall %d", f->type.host);
				write_asm(buf);
				if (!settings.nonet && (vp->attrs & (INPUT | OUTPUT)))
				{
					if (!vp->load_size)
						vp->load_size = 8;
					sprintf(buf, "pop%d a", vp->load_size);
					write_asm(buf);
//					write_asm("pop8 a");
					if (!vp->size)
					{
						write_asm("emode");
						write_asm("push a");
					}
					else
					{
						if (vp->attrs & STRING)
							write_asm("slen b,a");
						else if (vp->attrs & SYNC)
						{
							sprintf(buf, "ld b,%d", vp->size);
							write_asm(buf);
						}
						write_asm("rmode");
						write_asm("rsend a");
						if (vp->attrs & (STRING | SYNC))
						{
							write_asm("rsend b");
							write_asm("msend a,b");
						}
						write_asm("sync");
						write_asm("emode");
						write_asm("emode");
						sprintf(buf, "push%d a", vp->load_size);
						write_asm(buf);
//						write_asm("push8 a");
					}
				}
			}
			else
			{
				sprintf(buf, "call _%s", f->type.name);
				write_asm(buf);
			}
			memcpy(v, &f->type, sizeof(var));
			v->func = 0;
		}
		else
		{
			eval(str, &varg1, 0, 0);
			write_asm("pop a");
			if (varg1.attrs & HOST)
				write_asm("hcall a");
			else
				write_asm("call a");
			memcpy(v, &varg1, sizeof(var));
			v->fptr = 0;
		}

		v->args = NULL;

		return 0;
	}

	for (j = N_ORDERS; j; --j)
	{
		tflag = -1;
		status = 0;
		ap1 = (IS_ASN(j)) ? arg1 : arg2;
		ap2 = (IS_ASN(j)) ? arg2 : arg1;
		op = opr;
		paren = quote = squote = slash = 0;
		for (k = (IS_ASN(j)) ? 0 : sz-1;
			(IS_ASN(j)) ? (k < sz) : (k >= 0);
			(IS_ASN(j)) ? ++k : --k)
		{
			switch (status)
			{
			case 0:
				if (quote)
				{
					if (slash)
						slash = 0;
					else if (str[k] == '\\')
						slash = 1;
					else if (str[k] == '"')
						quote = 0;
					*ap1++ = str[k];
				}
				else if (squote)
				{
					if (slash)
						slash = 0;
					else if (str[k] == '\\')
						slash = 1;
					else if (str[k] == '\'')
						squote = 0;
					*ap1++ = str[k];
				}
				else if (paren)
				{
					if (IS_ASN(j))
					{
						if (str[k] == '(' || str[k] == '[')
							++paren;
						else if (str[k] == ')' || str[k] == ']')
							--paren;
					}
					else
					{
						if (str[k] == ')' || str[k] == ']')
							++paren;
						else if (str[k] == '(' || str[k] == '[')
							--paren;
					}
					*ap1++ = str[k];
				}
				else if (IS_ASN(j) ? (str[k] == '(' || str[k] == '[')
					: (str[k] == ')' || str[k] == ']'))
				{
					paren = 1;
					*ap1++ = str[k];
				}
				else if (is_opchar(str[k]))
				{
					*op++ = str[k];
					if (IS_ASN(j) ? k == sz-1 : !k)
						status = 2;
					else
						status = 1;
				}
				else
					*ap1++ = str[k];
				break;
			case 1:
				if (is_opchar(str[k]) || isspace(str[k]))
				{
					if (!isspace(str[k]))
						*op++ = str[k];
					if (IS_ASN(j) ? k == sz-1 : !k)
						status = 2;
				}
				else
				{
					*op = '\0';
					flag = 1;
					if (!strcmp(opr, "+") || !strcmp(opr, "-"))
					{
						if (IS_ASN(j))
						{
							if (*(ap1-1) == 'E' || *(ap1-1) == 'e')
							{
								for (sp = ap1-2; ; --sp)
									if ((is_opchar(*sp) &&
										*sp != '.') ||
										*sp == ' ')
										flag = 0;
									else if (!isdigit(*sp) &&
										*sp != '.')
										break;
							}
						}
						else
						{
							if (str[k] == 'E' || str[k] == 'e')
							{
								for (i = k-1; i >= 0; --i)
									if ((is_opchar(str[i]) &&
										str[i] != '.') ||
										str[i] == ' ')
										flag = 0;
									else if (!isdigit(str[i]) &&
										str[i] != '.')
										break;
							}
						}
					}
					if (flag)
					{
						for (n = k; (IS_ASN(j)) ? (n < sz) : (n >= 0);
							(IS_ASN(j)) ? ++n : --n)
							*ap2++ = str[n];
						*ap2 = '\0';
						status = 2;
					}
					else
					{
						if (IS_ASN(j))
						{
							*ap1++ = str[k];
							*ap1++ = *opr;
						}
						else
						{
							*ap1++ = *opr;
							*ap1++ = str[k];
						}
						op = opr;
						status = 0;
					}
				}
				break;
			}

			if (status == 2)
			{
				if (tflag == k)
					goto EOJ;
				tflag = k;
				*ap1 = '\0';
				*ap2 = '\0';
				*op = '\0';
				if (!IS_ASN(j))
				{
					_strrev(arg1);
					_strrev(arg2);
					_strrev(opr);
				}

				si = strlen(opr);
				for (i = 0; i < si; ++i)
				{
					on = -1;
					ons = 0;
					for (n = 0; n < N_OPERS; ++n)
						if (!strncmp(opers[n].op, &opr[i], strlen(opers[n].op))
							&& strlen(opers[n].op) > ons)
						{
							on = n;
							ons = strlen(opers[n].op);
						}
					sp = IS_ASN(j) ? arg1 : arg2;
					sp2 = IS_ASN(j) ? arg2 : arg1;
					if (on == DEC_POST && i == si-2 && *sp &&
						(isdigit(*sp2) || *sp2 == '.'))
						on = SUB;
					else if (on == INC_POST && i == si-2 && *sp &&
						(isdigit(*sp2) || *sp2 == '.'))
						on = ADD;

					if (on != -1)
					{
						for (i2 = 0; i2 < N_OPERS; ++i2)
							if (opers[i2].order == j &&
								!strcmp(opers[i2].op, opers[on].op))
								goto DoOp;
						i += strlen(opers[on].op)-1;
						on = -1;
					}
				}

DoOp:			if (on > -1)
				{
					n = on;

					if (n == COLON)
					{
						paren = quote = squote = slash = 0;
						for (sp = &arg1[strlen(arg1)-1]; ; --sp)
						{
							if (slash)
								slash = 0;
							else if (quote)
							{
								if (*sp == '\\')
									slash = 1;
								else if (*sp = '"')
									quote = 0;
							}
							else if (squote)
							{
								if (*sp == '\\')
									slash = 1;
								else if (*sp == '\'')
									squote = 0;
							}
							else if (*sp == '"')
								quote = 1;
							else if (*sp == '\'')
								squote = 1;
							else if (paren)
							{
								if (*sp == '(' || *sp == '[')
									++paren;
								else if (*sp == ')' || *sp == ']')
									--paren;
							}
							else if (*sp == '(' || *sp == '[')
								paren = 1;
							else if (*sp == '?')
							{
								*sp++ = '\0';
								strcpy(earg, arg1);
								strcpy(buf, sp);
								strcpy(arg1, buf);
								break;
							}
							if (sp == arg1)
							{
								PRINT_ERR("No `?' for `:'");
								return 1;
							}
						}
					}

					opr[i] = '\0';
					strcat(arg1, opr);
					sprintf(buf, "%s%s", &opr[i + strlen(opers[n].op)], arg2);
					strcpy(arg2, buf);

					if (n == PLUS && j == opers[ADD].order)
						n = ADD;
					else if (n == NEG && j == opers[SUB].order)
						n = SUB;
					else if (n == INDIR && j == opers[MULT].order)
						n = MULT;
					else if (n == ADDR && j == opers[AND].order)
						n = AND;
					else if (n == INC_POST && j == opers[INC_PRE].order)
						n = INC_PRE;
					else if (n == DEC_POST && j == opers[DEC_PRE].order)
						n = DEC_PRE;

					if (opers[n].order != j ||
						(opers[n].type == PREFIX && (arg1[0] || !arg2[0])) ||
						(opers[n].type == POSTFIX && (!arg1[0] || arg2[0])) ||
						(opers[n].type == BINARY && (!arg1[0] ||
						!arg2[0])))
							goto NoDice;

					if (n == NEG)
					{
						trim_spaces(arg2);
						if (is_int(arg2))
						{
							if (pp)
								return -atoi(arg2);
							else
							{
								sprintf(buf, "ld a,-%s", arg2);
								write_asm(buf);
								write_asm("push a");

								v->attrs = SIGNED | INT | CONST;
								set_sizes(v);
								return 0;
							}
						}
						else if (is_float(arg2))
						{
							if (pp)
								return -atoi(arg2);
							else
							{
								sprintf(buf, "ld8 a,-%s", arg2);
								write_asm(buf);
								write_asm("push8 a");

								v->attrs = CONST | DOUBLE;
								set_sizes(v);
								return 0;
							}
						}
					}

					if (n == DOT)
					{
						a1 = eval(arg1, &varg1, pp, 1);
						if (!(varg1.attrs & (STRUCT | UNION)))
						{
							sprintf(buf, "Left argument to `%s' not of `struct' or `union' type", opers[n].op);
							PRINT_ERR(buf);
							return 1;
						}
						else if (!(vp = find_stun_member(arg2, varg1.t.s)))
						{
							sprintf(buf, "`%s' not a member of left argument", arg2);
							PRINT_ERR(buf);
							return 1;
						}
					}
					else if (n == ARROW)
					{
						a1 = eval(arg1, &varg1, pp, 0);
						if (!(varg1.attrs & (STRUCT | UNION)))
						{
							sprintf(buf, "Left argument to `%s' not of `struct' or `union' type", opers[n].op);
							PRINT_ERR(buf);
							return 1;
						}
						else if (!(vp = find_stun_member(arg2, varg1.t.s)))
						{
							sprintf(buf, "`%s' not a member of left argument", arg2);
							PRINT_ERR(buf);
							return 1;
						}
					}
					else switch (opers[n].type)
					{
					case BINARY:
						a1 = eval(arg1, &varg1, pp, LVAL(n));
						a2 = eval(arg2, &varg2, pp, 0);

						if (LVAL(n) && (varg1.attrs & CONST))
						{
							PRINT_ERR("Cannot use `const' expression as lvalue");
							return 1;
						}

						if (IS_ASN(j))
						{
							--varg1.ptr;
							set_sizes(&varg1);
						}

						if (!pp && typecmp(&varg1, &varg2))
						{
							if (IS_ASN(j))
							{
								cast(&varg2, &varg1);
							}
							else if (varg1.ptr && !varg2.ptr)
							{
								cast(&varg2, &varg1);
								if (n == ADD || n == SUB ||
									n == ASN_ADD || n == ASN_SUB)
								{
									--varg1.ptr;
									set_sizes(&varg1);
									if (varg1.unit_size > 1)
									{
										write_asm("pop a");
										sprintf(buf, "ld b,%d",
											varg1.unit_size);
										write_asm(buf);
										write_asm("mult a,b");
										write_asm("push a");
									}
									++varg1.ptr;
									set_sizes(&varg1);
								}
							}
							else if (!varg1.ptr && varg2.ptr)
							{
								write_asm("pop a");
								cast(&varg1, &varg2);
								if (n == ADD || n == SUB)
								{
									--varg1.ptr;
									set_sizes(&varg1);
									if (varg1.unit_size > 1)
									{
										sprintf(buf, "pop%d a",
											varg2.load_size);
										write_asm(buf);
										sprintf(buf, "ld b,%d",
											varg1.unit_size);
										write_asm(buf);
										write_asm("mult a,b");
										sprintf(buf, "push%d a",
											varg2.load_size);
										write_asm(buf);
									}
									++varg1.ptr;
									set_sizes(&varg1);
								}
								write_asm("push a");
							}
							else
							{
								if (choose_cast(&varg1, &varg2) == 1)
								{
									sprintf(buf, "pop%d a",
										varg2.load_size);
									write_asm(buf);
									cast(&varg1, &varg2);
									sprintf(buf, "push%d a",
										varg2.load_size);
									write_asm(buf);
								}
								else
									cast(&varg2, &varg1);
							}
						}
						break;
					case PREFIX:
						a1 = eval(arg2, &varg1, pp, LVAL(n));

						if (n != ADDR && LVAL(n) && (varg1.attrs & CONST))
						{
							PRINT_ERR("Cannot use `const' expression as lvalue");
							return 1;
						}

						break;
					case POSTFIX:
						a1 = eval(arg1, &varg1, pp, LVAL(n));

						if (LVAL(n) && (varg1.attrs & CONST))
						{
							PRINT_ERR("Cannot use `const' expression as lvalue");
							return 1;
						}

						break;
					}

					if ((varg1.attrs & (FLOAT | DOUBLE)) && (
						n == ADD || n == SUB || n == MULT ||
						n == DIV || n == NEG || n == INC_PRE ||
						n == INC_POST || n == DEC_PRE ||
						n == DEC_POST || n == ASN_ADD ||
						n == ASN_SUB || n == ASN_MULT ||
						n == ASN_DIV || IS_TEST(j)))
						op_base[0] = 'F';
					else
						op_base[0] = '\0';

					if (pp)
					{
						if (varg1.attrs & (FLOAT | DOUBLE) ||
							varg2.attrs & (FLOAT | DOUBLE))
						{
							if (varg1.attrs & (FLOAT | DOUBLE))
								f1 = *(_float64 *)&a1;
							else
								f1 = a1;
							if (varg2.attrs & (FLOAT | DOUBLE))
								f2 = *(_float64 *)&a2;
							else
								f2 = a2;

							switch (n)
							{
							case ADD:
								f1 += f2;
								break;
							case SUB:
								f1 -= f2;
								break;
							case MULT:
								f1 *= f2;
								break;
							case DIV:
								f1 /= f2;
								break;
							case NEG:
								f1 = -f1;
								break;
							case LT:
								f1 = f1 < f2;
								break;
							case GT:
								f1 = f1 > f2;
								break;
							case LTE:
								f1 = f1 <= f2;
								break;
							case GTE:
								f1 = f1 >= f2;
								break;
							case EQ:
								f1 = f1 == f2;
								break;
							case NEQ:
								f1 = f1 != f2;
								break;
							case BAND:
								f1 = f1 && f2;
								break;
							case BOR:
								f1 = f1 || f2;
								break;
							case BNOT:
								f1 = !f1;
								break;
							case COMMA:
								f1 = f2;
								break;
							default:
								return 0;
							}

							v->attrs |= DOUBLE;
							return *(_int64 *)&f1;
						}
						else switch (n)
						{
						case ADD:
							return a1 + a2;
						case SUB:
							return a1 - a2;
						case MULT:
							return a1 * a2;
						case DIV:
							return a1 / a2;
						case NEG:
							return -a1;
						case MOD:
							return a1 % a2;
						case SLN:
							return a1 << a2;
						case SRN:
							return a1 >> a2;
						case AND:
							return a1 & a2;
						case XOR:
							return a1 ^ a2;
						case OR:
							return a1 | a2;
						case NOT:
							return ~a1;
						case LT:
							return a1 < a2;
						case GT:
							return a1 > a2;
						case LTE:
							return a1 <= a2;
						case GTE:
							return a1 >= a2;
						case EQ:
							return a1 == a2;
						case NEQ:
							return a1 != a2;
						case BAND:
							return a1 && a2;
						case BOR:
							return a1 || a2;
						case BNOT:
							return !a1;
						case COMMA:
							return a2;
						default:
							return 0;
						}
					}
					else
					{
						if (n == INDIR)
						{
							if (varg1.attrs & VOID)
								PRINT_ERR("`void' pointer invalid operand for unary `*'");
							else if (lvalue)
								set_sizes(&varg1);
							else
							{
								--varg1.ptr;
								set_sizes(&varg1);
								write_asm("pop b");
								sprintf(buf, "ld%d a,(b)",
									varg1.load_size);
								write_asm(buf);
								sprintf(buf, "push%d a",
									varg1.load_size);
								write_asm(buf);
							}
							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == ADDR)
						{
							if (lvalue)
								PRINT_ERR("Invalid lvalue");
							else
								memcpy(v, &varg1, sizeof(var));
						}
						else if (n == COLON)
						{
							eval(earg, &varg3, 0, 0);
							sprintf(buf, "pop%d a", varg3.load_size);
							write_asm(buf);
							sprintf(buf, "test%d a", varg3.load_size);
							write_asm(buf);
							sprintf(buf, "jp z,%s", lb1 = new_label());
							write_asm(buf);
							eval(arg1, &varg1, 0, 0);
							sprintf(buf, "jp %s", lb2 = new_label());
							write_asm(buf);
							sprintf(buf, "%s:", lb1);
							write_asm(buf);
							eval(arg2, &varg2, 0, 0);
							sprintf(buf, "%s:", lb2);
							write_asm(buf);

							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == DOT)
						{
							if (varg1.ptr != 1)
							{
								PRINT_ERR("Left argument of `.' has improper pointer properties");
								return 1;
							}

							memcpy(v, vp, sizeof(var));

							if (vp->loc && (varg1.attrs & STRUCT))
							{
								write_asm("pop a");
								sprintf(buf, "ld b,%d", vp->loc);
								write_asm(buf);
								write_asm("add a,b");
								write_asm("push a");
							}

							if (lvalue)
								++v->ptr;
							else if (!(v->arr_base && !v->argarr))
							{
								write_asm("pop a");
								if (v->bf_size == -1)
								{
									sprintf(buf, "ld%d b,(a)", vp->load_size);
									write_asm(buf);
									sprintf(buf, "push%d b", vp->load_size);
									write_asm(buf);
								}
								else
								{
									sprintf(buf, "; Loc: %d, Size: %d\n", v->bf_loc, v->bf_size);
									write_asm(buf);
									i = roundoff((float)(v->bf_loc + v->bf_size) / 8);
									if (i < 4)
										write_asm("ld b,0");
									sprintf(buf, "ld%d b,(a)", i);
									write_asm(buf);
									sprintf(buf, "%sgbf b,%d", (v->attrs & SIGNED) ? "s" : "",
										(v->bf_size << 8) + v->bf_loc);
									write_asm(buf);
									write_asm("push b");
								}
							}
						}
						else if (n == ARROW)
						{
							if (varg1.ptr != 1)
							{
								PRINT_ERR("Left argument of `->' has improper pointer properties");
								return 1;
							}

							memcpy(v, vp, sizeof(var));

							write_asm("pop a");
							if (vp->loc && (varg1.attrs & STRUCT))
							{
								sprintf(buf, "ld b,%d", vp->loc);
								write_asm(buf);
								write_asm("add a,b");
							}

							if (lvalue)
							{
								write_asm("push a");
								++v->ptr;
							}
							else if (!(v->arr_base && !v->argarr))
							{
								if (v->bf_size == -1)
								{
									sprintf(buf, "ld%d b,(a)", vp->load_size);
									write_asm(buf);
									sprintf(buf, "push%d b", vp->load_size);
									write_asm(buf);
								}
								else
								{
									sprintf(buf, "; Loc: %d, Size: %d\n", v->bf_loc, v->bf_size);
									write_asm(buf);
									i = roundoff((float)(v->bf_loc + v->bf_size) / 8);
									if (i < 4)
										write_asm("ld b,0");
									sprintf(buf, "ld%d b,(a)", i);
									write_asm(buf);
									sprintf(buf, "%sgbf b,%d", (v->attrs & SIGNED) ? "s" : "",
										(v->bf_size << 8) + v->bf_loc);
									write_asm(buf);
									write_asm("push b");
								}
							}

							print_var_info(v);
						}
						else if (n == ASN)
						{
							sprintf(buf, "pop%d b",
								varg2.load_size);
							write_asm(buf);
							write_asm("pop a");
							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (a),b",
									varg1.load_size);
								write_asm(buf);
								sprintf(buf, "push%d b",
									varg2.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((float)(varg1.bf_loc + varg1.bf_size) / 8);
								sprintf(buf, "ld%d c,(a)", i);
								write_asm(buf);
								sprintf(buf, "%ssbf c,b,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (a),c", i);
								write_asm(buf);
								sprintf(buf, "push%d b",
									varg2.load_size);
								write_asm(buf);
								varg1.bf_size = -1;
							}
							memcpy(v, &varg1, sizeof(var));
						}
						else if (IS_ASN(j))
						{
							sprintf(buf, "pop%d b",
								varg2.load_size);
							write_asm(buf);
							write_asm("pop c");
							sprintf(buf, "ld%d a,(c)",
								varg1.load_size);
							write_asm(buf);

							sprintf(buf, "%s%s%d a,b",
								op_base, &opers[n].cmd[4],
								varg1.load_size);
							write_asm(buf);

							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (c),a",
									varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((varg1.bf_loc + varg1.bf_size) / 8.0f);
								sprintf(buf, "ld%d b,(c)", i);
								write_asm(buf);
								sprintf(buf, "%ssbf b,a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (c),b", i);
								write_asm(buf);
								varg1.bf_size = -1;
							}

							sprintf(buf, "push%d a",
								varg2.load_size);
							write_asm(buf);
							memcpy(v, &varg1, sizeof(var));
						}
						else if (IS_TEST(j))
						{
							sprintf(buf, "pop%d b",
								varg2.load_size);
							write_asm(buf);
							sprintf(buf, "pop%d a",
								varg1.load_size);
							write_asm(buf);
							sprintf(buf, "%s%scp%d a,b",
								(varg1.attrs & SIGNED) ? "s" : "",
								op_base, varg1.load_size);
							write_asm(buf);
							sp = new_label();

							switch (n)
							{
							case EQ:
								write_asm("ld c,0");
								sprintf(buf, "jp nz,%s", sp);
								write_asm(buf);
								write_asm("ld c,1");
								break;
							case NEQ:
								write_asm("ld c,0");
								sprintf(buf, "jp z,%s", sp);
								write_asm(buf);
								write_asm("ld c,1");
								break;
							case GT:
								write_asm("ld c,0");
								sprintf(buf, "jp z,%s", sp);
								write_asm(buf);
								sprintf(buf, "jp c,%s", sp);
								write_asm(buf);
								write_asm("ld c,1");
								break;
							case GTE:
								write_asm("ld c,0");
								sprintf(buf, "jp c,%s", sp);
								write_asm(buf);
								write_asm("ld c,1");
								break;
							case LT:
								write_asm("ld c,0");
								sprintf(buf, "jp nc,%s", sp);
								write_asm(buf);
								write_asm("ld c,1");
								break;
							case LTE:
								write_asm("ld c,1");
								sprintf(buf, "jp c,%s", sp);
								write_asm(buf);
								sprintf(buf, "jp z,%s", sp);
								write_asm(buf);
								write_asm("ld c,0");
								break;
							}

							sprintf(buf, "%s:", sp);
							write_asm(buf);
							write_asm("push c");

							v->attrs = SIGNED | INT;
						}
						else if (n == INC_PRE)
						{
							--varg1.ptr;
							set_sizes(&varg1);
							write_asm("pop b");

							if (v->bf_size == -1)
							{
								sprintf(buf, "ld%d a,(b)", varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((float)(varg1.bf_loc + varg1.bf_size) / 8);
								if (i < 4)
									write_asm("ld a,0");
								sprintf(buf, "ld%d a,(b)", i);
								write_asm(buf);
								sprintf(buf, "%sgbf a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) + varg1.bf_loc);
								write_asm(buf);
							}

							sprintf(buf, "%sinc%d a",
								op_base, varg1.load_size);
							write_asm(buf);
							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (b),a",
									varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((varg1.bf_loc + varg1.bf_size) / 8.0f);
								sprintf(buf, "ld%d c,(b)", i);
								write_asm(buf);
								sprintf(buf, "%ssbf c,a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (b),c", i);
								write_asm(buf);
								if (!lvalue)
									varg1.bf_size = -1;
							}
							if (lvalue)
							{
								write_asm("push b");
								++varg1.ptr;
							}
							else
							{
								sprintf(buf, "push%d a",
									varg1.load_size);
								write_asm(buf);
							}
							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == DEC_PRE)
						{
							--varg1.ptr;
							set_sizes(&varg1);
							write_asm("pop b");

							if (v->bf_size == -1)
							{
								sprintf(buf, "ld%d a,(b)", varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((float)(varg1.bf_loc + varg1.bf_size) / 8);
								if (i < 4)
									write_asm("ld a,0");
								sprintf(buf, "ld%d a,(b)", i);
								write_asm(buf);
								sprintf(buf, "%sgbf a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) + varg1.bf_loc);
								write_asm(buf);
							}
							
							sprintf(buf, "%sdec%d a",
								op_base, varg1.load_size);
							write_asm(buf);

							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (b),a",
									varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((varg1.bf_loc + varg1.bf_size) / 8.0f);
								sprintf(buf, "ld%d c,(b)", i);
								write_asm(buf);
								sprintf(buf, "%ssbf c,a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (b),c", i);
								write_asm(buf);
								if (!lvalue)
									varg1.bf_size = -1;
							}

							sprintf(buf, "push%d a",
								varg1.load_size);
							write_asm(buf);
							if (lvalue)
							{
								write_asm("push b");
								++varg1.ptr;
							}
							else
							{
								sprintf(buf, "push%d a",
									varg1.load_size);
								write_asm(buf);
							}
							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == INC_POST)
						{
							--varg1.ptr;
							set_sizes(&varg1);
							write_asm("pop b");

							if (v->bf_size == -1)
							{
								sprintf(buf, "ld%d a,(b)", varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((float)(varg1.bf_loc + varg1.bf_size) / 8);
								if (i < 4)
									write_asm("ld a,0");
								sprintf(buf, "ld%d a,(b)", i);
								write_asm(buf);
								sprintf(buf, "%sgbf a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) + varg1.bf_loc);
								write_asm(buf);
							}

							if (lvalue)
							{
								write_asm("push b");
								++varg1.ptr;
							}
							else
							{
								sprintf(buf, "push%d a",
									varg1.load_size);
								write_asm(buf);
							}
							sprintf(buf, "%sinc%d a",
								op_base, varg1.load_size);
							write_asm(buf);

							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (b),a",
									varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((varg1.bf_loc + varg1.bf_size) / 8.0f);
								sprintf(buf, "ld%d c,(b)", i);
								write_asm(buf);
								sprintf(buf, "%ssbf c,a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (b),c", i);
								write_asm(buf);
								if (!lvalue)
									varg1.bf_size = -1;
							}

							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == DEC_POST)
						{
							--varg1.ptr;
							set_sizes(&varg1);
							write_asm("pop b");

							if (v->bf_size == -1)
							{
								sprintf(buf, "ld%d a,(b)", varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((float)(varg1.bf_loc + varg1.bf_size) / 8);
								if (i < 4)
									write_asm("ld a,0");
								sprintf(buf, "ld%d a,(b)", i);
								write_asm(buf);
								sprintf(buf, "gbf a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) + varg1.bf_loc);
								write_asm(buf);
							}

							if (lvalue)
							{
								write_asm("push b");
								++varg1.ptr;
							}
							else
							{
								sprintf(buf, "push%d a",
									varg1.load_size);
								write_asm(buf);
							}
							sprintf(buf, "%sdec%d a",
								op_base, varg1.load_size);
							write_asm(buf);

							if (varg1.bf_size == -1)
							{
								sprintf(buf, "ld%d (b),a",
									varg1.load_size);
								write_asm(buf);
							}
							else
							{
								i = roundoff((varg1.bf_loc + varg1.bf_size) / 8.0f);
								sprintf(buf, "ld%d c,(b)", i);
								write_asm(buf);
								sprintf(buf, "%sbf c,a,%d",
									(v->attrs & SIGNED) ? "s" : "",
									(varg1.bf_size << 8) +
									varg1.bf_loc);
								write_asm(buf);
								sprintf(buf, "ld%d (b),c", i);
								write_asm(buf);
								if (!lvalue)
									varg1.bf_size = -1;
							}

							memcpy(v, &varg1, sizeof(var));
						}
						else if (n == COMMA)
						{
							sprintf(buf, "pop%d a",
								varg2.load_size ? varg2.load_size :
								1);
							write_asm(buf);
							write_asm("pop b");
							sprintf(buf, "push%d a",
								varg2.load_size ? varg2.load_size :
								1);
							write_asm(buf);
							memcpy(v, &varg2, sizeof(var));
						}
						else
						{
							switch (opers[n].type)
							{
							case BINARY:
								sprintf(buf, "pop%d b", varg1.load_size);
								write_asm(buf);
								sprintf(buf, "pop%d a", varg1.load_size);
								write_asm(buf);
								sprintf(buf, "%s%s%s%d a,b", op_base,
									opers[n].cmd, ((n == MULT || n == DIV)
									&& (varg1.attrs & SIGNED)) ? "s" : "", varg1.load_size);
								write_asm(buf);
								sprintf(buf, "push%d a", varg1.load_size);
								write_asm(buf);
								break;
							case PREFIX:
							case POSTFIX:
								if (n != PLUS)
								{
									sprintf(buf, "pop%d a",
										varg1.load_size);
									write_asm(buf);
									sprintf(buf, "%s%s%d a", op_base,
										opers[n].cmd, varg1.load_size);
									write_asm(buf);
									sprintf(buf, "push%d a",
										varg1.load_size);
									write_asm(buf);
								}
							}
							memcpy(v, &varg1, sizeof(var));
						}

						set_sizes(v);
						return 0;
					}
				}
NoDice:			if (IS_ASN(j))
				{
					strcat(arg1, opr);
					ap2 = arg2;
					--k;
				}
				else
				{
					strcpy(sword, arg2);
					strcpy(arg2, opr);
					strcat(arg2, sword);
					_strrev(arg2);
					ap2 = arg1;
					++k;
				}
				ap1 += strlen(opr);
				op = opr;
				status = 0;
			}
		}
EOJ:;
	}

	PRINT_ERR("Syntax error");
	puts(str);
	errno = 1;
	return 1;
}

int preprocess(file_info *dfile, file_info *sfile, char *src,
			   int *pos, int *osz)
{
	char quote = 0, slash = 0, nl = 1, squote = 0, cmnt = 0;
	char if_cmnt = 0;
	int paren = 0, ifs = 0, ifsz = 0, paren2, if_break = -1;
	char ch, copy, wend, word[WORD_LENGTH], sword[WORD_LENGTH], *wp;
	char cmd[WORD_LENGTH], args[WORD_LENGTH], *cp, *ap, space;
	int j, sj, k, sk, errs = 0, line = 1, n;
	char *out = NULL, pbuf[PATH_LENGTH], *arg2, *arg3;
	file_info finf, finf2, *fi;
	macro *m, *m2;
	lstr *ls, *ls_base;
	var varg;
	char *ifstat = (char *)malloc(0), cflag;
	char err[WORD_LENGTH];
	pevent *pe;

	if (++inc_level == INC_MAX)
	{
		sprintf(err, "Maximum level of %d nested `#includes' reached; aborting", INC_MAX);
		print_err(dfile, *pos, err);
		return 1;
	}
	else if (inc_level > INC_MAX)
		return 1;

	finf.text = NULL;
	finf2.text = NULL;

	wp = word;

	if (dfile->n_lines)
		if (dfile->lines[dfile->n_lines-1].pos == *pos)
		{
			dfile->lines[dfile->n_lines-1].line = 1;
			dfile->lines[dfile->n_lines-1].file = sfile;
			goto SkipIt;
		}
	dfile->lines = (line_info *)realloc(dfile->lines,
		++dfile->n_lines * sizeof(line_info));
	dfile->lines[dfile->n_lines-1].line = 1;
	dfile->lines[dfile->n_lines-1].pos = *pos;
	dfile->lines[dfile->n_lines-1].file = sfile;
SkipIt:

	sj = strlen(src);
	for (j = 0; j < sj; ++j)
	{
		ch = src[j];

		wend = 0;
		if (cmnt)
		{
			if (ch == '*' && src[j+1] == '/')
			{
				cmnt = 0;
				++j;
			}
			copy = 0;
		}
		else if (ch == '/' && src[j+1] == '/')
		{
			for (j += 2; j < sj && (ch = src[j]) != '\n'; ++j);
			copy = 2;
		}
		else if (ch == '/' && src[j+1] == '*')
		{
			cmnt = 1;
			++j;
			copy = 0;
			if (*pos >= *osz)
			{
				*osz += BLOCK_SIZE;
				dfile->text = (char *)realloc(dfile->text, *osz * sizeof(char));
			}
			dfile->text[(*pos)++] = ' ';
		}
		else if (nl && ch == '#')
		{
			space = 0;
			*(cp = cmd) = '\0';
			*(ap = args) = '\0';
			for (++j; isspace(src[j]) && src[j] != '\n'; ++j);
			for (; ch = src[j], j < sj && !(ch == '\n' && src[j-1] != '\\'); ++j)
			{
				if (ch == '\n' && src[j-1] == '\\')
				{
					if (space)
						--ap;
					else
						--cp;
				}
				else if (space)
					*ap++ = ch;
				else if (isspace(ch))
					space = 1;
				else
					*cp++ = ch;
			}
			*cp = '\0';
			*ap = '\0';
			trim_spaces(cmd);
			trim_spaces(args);

			cflag = 0;
			if (!strcmp(cmd, "ifdef"))
			{
				cflag = 1;
				if (!if_cmnt)
				{
					if (if_cmnt = !macro_defined(args))
						if_break = ifs;
					else
						if_break = -1;
				}
				if (++ifs > ifsz)
					ifstat = (char *)realloc(ifstat, (ifsz = ifs) *
						sizeof(char));
				ifstat[ifs-1] = macro_defined(args);
			}
			else if (!strcmp(cmd, "ifndef"))
			{
				cflag = 1;
				if (!if_cmnt)
				{
					if (if_cmnt = macro_defined(args))
						if_break = ifs;
					else
						if_break = -1;
				}
				if (++ifs > ifsz)
					ifstat = (char *)realloc(ifstat, (ifsz = ifs) *
						sizeof(char));
				ifstat[ifs-1] = !macro_defined(args);
			}
			else if (!strcmp(cmd, "if"))
			{
				um_base = NULL;
				while (replace_macros(args, sword, 1, sfile->name, line))
					strcpy(args, sword);
				free_lstr2(um_base);
				strcpy(args, sword);

				cflag = 1;
				if (!if_cmnt)
				{
					if (if_cmnt = !(n = eval(args, &varg, 1, 0)))
						if_break = ifs;
					else
						if_break = -1;
				}
				if (++ifs > ifsz)
					ifstat = (char *)realloc(ifstat, (ifsz = ifs) *
						sizeof(char));
				ifstat[ifs-1] = n;
			}
			else if (!strcmp(cmd, "elif"))
			{
				um_base = NULL;
				while (replace_macros(args, sword, 1, sfile->name, line))
					strcpy(args, sword);
				free_lstr2(um_base);
				strcpy(args, sword);

				cflag = 1;
				if (!ifs)
					print_err(dfile, *pos, "#elif without #if");
				else if (if_break == -1 || ifs-1 == if_break)
				{
					if (ifstat[ifs-1])
						if_cmnt = 1;
					else if (eval(args, &varg, 1, 0))
					{
						if_cmnt = 0;
						ifstat[ifs-1] = 1;
					}
					if (if_cmnt)
						if_break = ifs-1;
					else
						if_break = -1;
				}
			}
			else if (!strcmp(cmd, "else"))
			{
				cflag = 1;
				if (!ifs)
					print_err(dfile, *pos, "#else without #if");
				else if (if_break == -1 || ifs-1 == if_break)
				{
					if_cmnt = ifstat[ifs-1];
					ifstat[ifs-1] = 1;
					if (if_cmnt)
						if_break = ifs-1;
					else
						if_break = -1;
				}
			}
			else if (!strcmp(cmd, "endif"))
			{
				cflag = 1;
				if (!ifs)
					print_err(dfile, *pos, "#endif without #if");
				else if (--ifs <= if_break)
				{
					if_cmnt = 0;
					if_break = -1;
				}
			}
			else if (if_cmnt)
				goto PP1;

			if (!strcmp(cmd, "include"))
			{
				fi = NULL;
				if (strlen(args) < 2)
					print_err(dfile, *pos, "Bad parameter for #include");
				else if (args[0] == '"' &&
					args[strlen(args)-1] == '"')
				{
					args[strlen(args)-1] = '\0';
					sprintf(pbuf, "%s/%s", path, &args[1]);
					if (!(fi = fetch_file(pbuf, &args[1])))
					{
						sprintf(err, "Unable to #include file \"%s\"\n",
							&args[1]);
						print_err(dfile, *pos, err);
					}
				}
				else if (args[0] == '<' &&
					args[strlen(args)-1] == '>')
				{
					args[strlen(args)-1] = '\0';
					sprintf(pbuf, "%s/%s", path, &args[1]);
					if (!(fi = fetch_file(pbuf, &args[1])))
					{
						sprintf(err, "Unable to #include file \"%s\"",
							&args[1]);
						print_err(dfile, *pos, err);
					}
				}
				else
					print_err(dfile, *pos, "Bad parameter for #include");

				if (fi)
				{
					errs += preprocess(dfile, fi,
						fi->text, pos, osz);
				}
			}
			else if (!strcmp(cmd, "define"))
			{
				sk = strlen(args);
				paren2 = 0;
				ls_base = NULL;
				for (k = 0; k <= sk; ++k)
				{
					if (paren2)
					{
						if (args[k] == ')')
						{
							args[k] = '\0';
							arg2 = &args[k+1];

							strcpy(sword, arg3);
							trim_spaces(sword);
							ls = (lstr *)calloc(1, sizeof(lstr));
							ls->str = (char *)calloc(strlen(sword)+1,
								sizeof(char));
							strcpy(ls->str, sword);
							ls->prev = ls_base;
							if (ls_base)
								ls_base->next = ls;
							ls_base = ls;

							break;
						}
						else if (args[k] == ',')
						{
							args[k] = '\0';
							strcpy(sword, arg3);
							trim_spaces(sword);
							ls = (lstr *)calloc(1, sizeof(lstr));
							ls->str = (char *)calloc(strlen(sword)+1,
								sizeof(char));
							strcpy(ls->str, sword);
							ls->prev = ls_base;
							if (ls_base)
								ls_base->next = ls;
							ls_base = ls;

							arg3 = &args[k+1];
						}
					}
					else if (args[k] == '(')
					{
						paren2 = 1;
						args[k] = '\0';
						arg3 = &args[k]+1;
					}
					else if (isspace(args[k]) || k == sk)
					{
						args[k] = '\0';
						arg2 = &args[k+1];
						break;
					}
				}

				if (!args[0])
					print_err(dfile, *pos, "No macro name specified for #define");
				else
				{
					trim_spaces(arg2);

					if (!macro_defined(args))
					{
						m = (macro *)calloc(1, sizeof(macro));

						for (m2 = macro_base; m2; m2 = m2->next)
							if (!m2->next)
								break;
						for (; m2; m2 = m2->prev)
						{
							if (!strncmp(args, m2->name,
								strlen(args)))
								break;
						}

						if (m2)
						{
							m->prev = m2;
							if (m2)
							{
								m->next = m2->next;
								if (m2->next)
									m2->next->prev = m;
								m2->next = m;
							}
						}
						else
						{
							m->next = macro_base;
							if (macro_base)
								macro_base->prev = m;
							macro_base = m;
						}
						m->name = (char *)calloc(strlen(args)+1,
							sizeof(char));
						strcpy(m->name, args);
					}
					else
					{
						sprintf(err, "Redefinition of macro %s\n",
							args);
						print_warn(dfile, *pos, err);
						free(m->value);
						free_lstr(m->args);
					}

					for (; ls_base; ls_base = ls_base->prev)
						if (!ls_base->prev)
							break;
					m->args = ls_base;
					m->value = (char *)calloc(strlen(arg2)+1,
						sizeof(char));
					strcpy(m->value, arg2);
				}
			}
			else if (!strcmp(cmd, "undef"))
			{
				if (!args[0])
				{
					puts("No macro name specified for #undef");
					++errs;
				}
				else
				{
					char flag = 1;
					for (m = macro_base; m; m = m->next)
					{
						if (!strcmp(m->name, args))
						{
							if (m->prev)
								m->prev->next = m->next;
							if (m->next)
								m->next->prev = m->prev;
							if (m == macro_base)
								macro_base = m->next;
							flag = 0;
							break;
						}
					}
					
					if (flag)
					{
						sprintf(err, "Attempt to #undef undefined macro %s\n",
							args);
						print_warn(dfile, *pos, err);
					}
				}
			}
			else if (!strcmp(cmd, "error"))
			{
				print_err(dfile, *pos, args);
			}
			else if (!strcmp(cmd, "pragma"))
			{
				pe = (pevent *)calloc(1, sizeof(pevent));
				strcpy(pe->str, args);
				pe->pos = *pos;
				pe->prev = cur_pevent;
				if (pevent_base)
					cur_pevent->next = pe;
				else
					pevent_base = pe;
				cur_pevent = pe;
			}
			else if (!strcmp(cmd, "line"))
			{
				um_base = NULL;
				while (replace_macros(args, sword, 0, sfile->name, line))
					strcpy(args, sword);
				free_lstr2(um_base);
				strcpy(args, sword);

				sk = strlen(args);
				arg2 = NULL;
				for (k = 0; k < sk; ++k)
				{
					if (isspace(args[k]))
					{
						args[k] = '\0';
						arg2 = &args[k+1];
						break;
					}
				}
				trim_spaces(arg2);
				n = atoi(args)-1;

				if (n < 1)
					print_err(dfile, *pos, "Invalid line number specified");
				else
				{
					line = n;
					if (arg2)
					{
						arg2[strlen(arg2)-1] = '\0';
						++arg2;
					}
					else
						arg2 = sfile->name;

					sfile = fetch_file(NULL, arg2);
					sfile->n_lines = 1;
					sfile->lines = (line_info *)calloc(1,
						sizeof(line_info));
					sfile->lines[0].file = sfile;
					sfile->lines[0].pos = *pos;
					sfile->lines[0].line = line;
				}
			}
			else if (cmd[0] && !cflag)
			{
				puts("Unknown preprocessor directive");
				++errs;
			}

			copy = 0;
		}
		else if (if_cmnt)
			copy = 0;
		else if (paren)
		{
			copy = 1;
			if (quote)
			{
				if (slash)
					slash = 0;
				else if (ch == '\\')
					slash = 1;
				else if (ch == '"')
					quote = 0;
			}
			else if (squote)
			{
				if (slash)
					slash = 0;
				else if (ch == '\\')
					slash = 1;
				else if (ch == '\'')
					squote = 0;
			}
			else if (ch == '"')
				quote = 1;
			else if (ch == '\'')
				squote = 1;
			else if (ch == '(')
				++paren;
			else if (ch == ')')
				--paren;
		}
		else if (quote)
		{
			if (slash)
				slash = 0;
			else if (ch == '\\')
				slash = 1;
			else if (ch == '"')
				quote = 0;
			copy = 2;
		}
		else if (squote)
		{
			if (slash)
				slash = 0;
			else if (ch == '\\')
				slash = 1;
			else if (ch == '\'')
				squote = 0;
			copy = 2;
		}
		else if (ch == '"')
		{
			quote = 1;
			copy = 2;
			wend = 1;
		}
		else if (ch == '\'')
		{
			squote = 1;
			copy = 2;
			wend = 1;
		}
		else if (isspace(ch))
		{
			copy = 2;
			wend = 1;
		}
		else if (ch == '(')
		{
			paren = 1;
			copy = 1;
		}
		else if (ch != '\r')
			copy = 1;

		if (wend)
		{
			if (wp != word)
			{
				*wp = '\0';
				um_base = NULL;
				while (replace_macros(word, sword, 0, sfile->name, line))
					strcpy(word, sword);
				free_lstr2(um_base);
				if (*pos + strlen(sword) >= *osz)
				{
					*osz += BLOCK_SIZE;
					dfile->text = (char *)realloc(dfile->text,
						*osz * sizeof(char));
				}
				strcpy(&dfile->text[*pos], sword);
				*pos += strlen(sword);
				wp = word;
			}
		}

PP1:	switch (copy)
		{
		case 1:
			*wp++ = ch;
			break;
		case 2:
			if (*pos >= *osz)
			{
				*osz += BLOCK_SIZE;
				dfile->text = (char *)realloc(dfile->text, 
					*osz * sizeof(char));
			}
			dfile->text[(*pos)++] = ch;
			break;
		}

		if (ch == '\n')
		{
			if (!quote && !squote)
			{
				switch (copy)
				{
				case 1:
					if (*(wp-2) == '\\')
					{
						wp -= 2;
						*wp++ = ' ';
						nl = 0;
					}
					else
						nl = 1;
					break;
				case 2:
					if (dfile->text[*pos-2] == '\\')
					{
						pos -= 2;
						*pos++ = ' ';
						nl = 0;
					}
					else
						nl = 1;
					break;
				}
			}
			else
				nl = 0;

			dfile->lines = (line_info *)realloc(dfile->lines,
				++dfile->n_lines * sizeof(line_info));
			dfile->lines[dfile->n_lines-1].line = ++line;
			dfile->lines[dfile->n_lines-1].pos = *pos;
			dfile->lines[dfile->n_lines-1].file = sfile;
		}
		else if (!isspace(ch))
			nl = 0;
	}

	free(ifstat);
	--inc_level;
	return errs;
}

#undef PRINT_ERR
#define PRINT_ERR(X) print_err(fi, line_pos, (X))

code *new_code(code **cur_code, code **code_base, char *linetxt,
			   lstr *label_base, code *needs_child, code **peer)
{
	code *c = (code *)calloc(1, sizeof(code));
	char flag = 0;

	if (peer)
		if (*peer)
			flag = 1;

	if (flag)
	{
		(*peer)->peer = c;
		c->pparent = *peer;
		*peer = (*peer)->cparent;
	}
	else if (needs_child)
	{
		needs_child->child = c;
		c->cparent = needs_child;
	}
	else
	{
		if (*cur_code)
			(*cur_code)->next = c;
		c->prev = *cur_code;

		if (!*code_base)
			*code_base = c;
		*cur_code = c;
	}

	c->label_base = label_base;

	if (linetxt)
	{
		c->str = (char *)calloc(strlen(linetxt)+1, sizeof(char));
		strcpy(c->str, linetxt);
	}

	return c;
}

void compile(file_info *fi, char **sp, int *pos, int *line,
			 int *next_line, code *parent)
{
	int j, paren = 0, brace = 0;
	char declr = 0, quote = 0, squote = 0, slash = 0, eop = 0;
	char sword[WORD_LENGTH], linetxt[WORD_LENGTH], *lp = linetxt;
	char dof = 0, wf = 0, ch, paren2, *sp2, *sw, newstat = 1;
	char dof2 = 0, buf[WORD_LENGTH], td = 0;
	char space, casef, quote2, squote2, slash2, flag;
	int _paren;
	code *c, *cur_code = NULL, *needs_child = NULL, *last_peer = NULL;
	lstr *ls, *label_base = NULL;
	var varg;

	for (; (ch = **sp); ++*sp)
	{
		if (++*pos > *next_line)
		{
			line_pos = *pos;
			cur_file = fi->lines[++*line].file;
			cur_line = fi->lines[*line].line;
			if (fi->n_lines > *line+1)
				*next_line = fi->lines[*line+1].pos;
			else
				*next_line = -1;
		}

		if (newstat)
		{
			sw = sword;
			label_base = NULL;
			space = casef = quote2 = squote2 = slash2 = _paren = 0;
			for (sp2 = *sp; *sp2; ++sp2)
				if (!isspace(*sp2))
					break;
			for (; *sp2; ++sp2)
			{
				if (quote2)
				{
					if (slash2)
						slash2 = 0;
					else if (*sp2 == '\\')
						slash2 = 1;
					else if (*sp2 == '"')
						quote2 = 0;
					*sw++ = *sp2;
				}
				else if (squote2)
				{
					if (slash2)
						slash2 = 0;
					else if (*sp2 == '\\')
						slash2 = 1;
					else if (*sp2 == '\'')
						squote2 = 0;
					*sw++ = *sp2;
				}
				else if (_paren)
				{
					if (*sp2 == '(' || *sp2 == '[')
						++_paren;
					else if (*sp2 == ')' || *sp2 == ']')
						--_paren;
					*sw++ = *sp2;
				}
				else if (*sp2 == ':')
				{
					*sw = '\0';
					trim_spaces(sword);

					ls = (lstr *)calloc(1, sizeof(lstr));
					ls->next = label_base;
					if (label_base)
						label_base->prev = ls;
					label_base = ls;

					ls->str = (char *)calloc(strlen(sword)+1,
						sizeof(char));
					strcpy(ls->str, sword);

					*pos += sp2 - *sp;
					*sp = sp2;
					sw = sword;
					sword[0] = '\0';
					space = casef = 0;
				}
				else if (casef && *sp2 == '"')
				{
					quote2 = 1;
					*sw++ = '"';
				}
				else if (casef && *sp2 == '\'')
				{
					squote2 = 1;
					*sw++ = '\'';
				}
				else if (casef && (*sp2 == '(' || *sp2 == '['))
				{
					paren2 = 1;
					*sw++ = *sp2;
				}
				else if (!casef && isspace(*sp2))
				{
					if (!strncmp(sword, "case", 4))
						casef = 1;
					*sw++ = *sp2;
					space = 1;
				}
				else if (!casef && !isspace(*sp2) && *sp2 != '_' &&
					!isalnum(*sp2))
					break;
				else
				{
					if (isspace(*sp2))
						space = 1;
					else if (!casef && space)
						break;

					*sw++ = *sp2;
				}
			}
			newstat = 0;
		}

		if (*pos > *next_line)
		{
			line_pos = *pos;
			cur_file = fi->lines[++*line].file;
			cur_line = fi->lines[*line].line;
			if (fi->n_lines > *line+1)
				*next_line = fi->lines[*line+1].pos;
			else
				*next_line = -1;
		}

		if (quote)
		{
			*lp++ = ch;
			if (slash)
				slash = 0;
			else if (ch == '\\')
				slash = 1;
			else if (ch == '"')
				quote = 0;
		}
		else if (squote)
		{
			*lp++ = ch;
			if (slash)
				slash = 0;
			else if (ch == '\\')
				slash = 1;
			else if (ch == '\'')
				squote = 0;
		}
		else if (ch == '"')
		{
			*lp++ = '"';
			quote = 1;
		}
		else if (ch == '\'')
		{
			*lp++ = '\'';
			squote = 1;
		}
		else if (declr)
		{
			if (quote)
			{
				*lp++ = ch;
				if (slash)
					slash = 0;
				else if (ch == '\\')
					slash = 1;
				else if (ch == '"')
					quote = 0;
			}
			else if (squote)
			{
				*lp++ = ch;
				if (slash)
					slash = 0;
				else if (ch == '\\')
					slash = 1;
				else if (ch == '\'')
					squote = 0;
			}
			else if (ch == '"')
			{
				*lp++ = '"';
				quote = 1;
			}
			else if (ch == '\'')
			{
				*lp++ = '\'';
				squote = 1;
			}
			else if (paren)
			{
				if (ch == '(' || ch == '[')
				{
					*lp++ = ch;
					++paren;
				}
				else if (ch == ')' || ch == ']')
				{
					*lp++ = ch;
					--paren;
				}
				else if (isspace(ch))
				{
					if (lp == linetxt ? 0 : !isspace(*(lp-1)))
						*lp++ = ' ';
				}
				else
					*lp++ = ch;
			}
			else if (brace)
			{
				if (ch == '{')
				{
					*lp++ = ch;
					++brace;
				}
				else if (ch == '}')
				{
					*lp++ = ch;
					--brace;
				}
				else if (isspace(ch))
				{
					if (lp == linetxt ? 0 : !isspace(*(lp-1)))
						*lp++ = ' ';
				}
				else
					*lp++ = ch;
			}
			else if (ch == '(' || ch == '[')
			{
				paren = 1;
				paren2 = (ch == '(');
				*lp++ = ch;
			}
			else if (ch == '{')
			{
				if (paren2)
				{
					declr = 0;

					flag = 0;
					if (parent == global.code_base)
						for (sp2 = *sp-1; *sp2; --sp2)
						{
							if (*sp2 == ')')
							{
								flag = 1;
								break;
							}
							else if (!isspace(*sp2))
								break;
						}

					if (flag)
					{
						newstat = 1;
						*lp = '\0';
						c = new_code(&cur_code, &parent->child,
							linetxt, label_base, NULL, 0);
						label_base = NULL;

						c->file = fi;
						c->pos = line_pos;
						c->type = FUNC;
						c->parent = parent;

						lp = linetxt;

						c = new_code(&cur_code, &parent->child,
							NULL, NULL, NULL, 0);
						c->type = MULTI;
						c->parent = parent;
						++*sp;

						compile(fi, sp, pos, line, next_line, c);
					}
					else
						PRINT_ERR("Parse error at '{'");
				}
				else
				{
					brace = 1;
					*lp++ = '{';
				}
			}
			else if (ch == ';')
			{
				newstat = 1;
				*lp = '\0';
				c = new_code(&cur_code, &parent->child, linetxt,
					label_base, NULL, 0);
				label_base = NULL;

				c->file = fi;
				c->pos = line_pos;
				c->type = DECLR;
				c->parent = parent;

				if (!strcmp(first_word(buf, linetxt), "typedef"))
					parse_declr(linetxt, &varg, NULL, NULL, 0);

				lp = linetxt;
				declr = 0;
				paren2 = 0;
			}
			else if (isspace(ch))
			{
				if (lp == linetxt ? 0 : !isspace(*(lp-1)))
					*lp++ = ' ';
			}
			else
				*lp++ = ch;
		}
		else if (paren)
		{
			if (ch == '(' || ch == '[')
			{
				*lp++ = ch;
				++paren;
			}
			else if (ch == ')' || ch == ']')
			{
				*lp++ = ch;
				if (!--paren && eop)
				{
					newstat = 1;
					eop = 0;
					*lp = '\0';
					c = new_code(&cur_code, &parent->child,
						linetxt, label_base, needs_child, 0);
					label_base = NULL;
					if (!strcmp(first_word(buf, c->str), "if"))
						last_peer = c;

					c->file = fi;
					c->pos = line_pos;
					c->type = CMND;
					c->parent = parent;

					lp = linetxt;
					needs_child = c;
				}
			}
			else if (isspace(ch))
			{
				if (lp == linetxt ? 0 : !isspace(*(lp-1)))
					*lp++ = ' ';
			}
			else
				*lp++ = ch;
		}
		else if (isspace(ch) || ch == '(' || ch == '*')
		{
			*lp = '\0';
			strcpy(sword, linetxt);
			trim_spaces(sword);
			if (ch != '*' && dof && !strcmp(sword, "while"))
			{
				dof = 0;
				dof2 = 1;
				wf = 1;
			}
			else if (ch != '*' && !strcmp(sword, "if") || !strcmp(sword, "for")
				|| (!wf && !strcmp(sword, "while")) ||
				!strcmp(sword, "switch"))
				eop = 1;
			else if (ch != '*' && !strcmp(sword, "else") || !strcmp(sword, "do"))
			{
				newstat = 1;

				if (!strcmp(sword, "do"))
				{
					dof = 1;
					c = new_code(&cur_code, &parent->child, linetxt,
						label_base, needs_child, NULL);
				}
				else
					c = new_code(&cur_code, &parent->child, linetxt,
						label_base, NULL, &last_peer);

				label_base = NULL;

				c->file = fi;
				c->pos = line_pos;
				c->type = CMND;
				c->parent = parent;

				lp = linetxt;
				needs_child = c;
				if (dof)
					last_peer = c;
			}
			else
			{
				if (parent == global.code_base)
					declr = 1;
				else for (j = 0; j < N_ATTRS; ++j)
					if (!strcmp(sword, attrs[j]))
					{
						declr = 1;
						paren = paren2 = brace = 0;
						break;
					}
				if (!declr)
					for (ls = temp_td_base; ls; ls = ls->next)
						if (!strcmp(ls->str, sword))
							{
								declr = 1;
								paren = paren2 = brace = 0;
							}
			}

			if (ch == '(')
			{
				++paren;
				*lp++ = '(';
			}
			else if (ch == '*')
			{
				*lp++ = '*';
			}
			else if (lp == linetxt ? 0 : !isspace(*(lp-1)))
				*lp++ = ' ';
		}
		else if (ch == ';')
		{
			newstat = 1;
			*lp = '\0';
			if (dof2)
				c = new_code(&cur_code, &parent->child, linetxt,
					label_base, needs_child, &last_peer);
			else
				c = new_code(&cur_code, &parent->child, linetxt,
					label_base, needs_child, NULL);

			label_base = NULL;

			c->file = fi;
			c->pos = line_pos;
			c->type = CMND;
			c->parent = parent;

			lp = linetxt;
			wf = dof2 = 0;
			needs_child = NULL;
		}
		else if (ch == '{')
		{
			newstat = 1;
			*lp = '\0';
			if (linetxt[0] && linetxt[1] && strcmp(linetxt, " "))
				PRINT_ERR("Extra odd code");

			c = new_code(&cur_code, &parent->child, NULL,
				label_base, needs_child, 0);
			label_base = NULL;

			c->type = MULTI;
			c->parent = parent;
			++*sp;

			needs_child = NULL;
			compile(fi, sp, pos, line, next_line, c);
		}
		else if (ch == '}')
		{
			if (parent == global.code_base)
				PRINT_ERR("Parse error at `}'");
			else if (linetxt[0] && linetxt[1] &&
				strcmp(linetxt, " "))
				PRINT_ERR("Parse error at `}'; possible unterminated statement");
			else
				return;
		}
		else
			*lp++ = ch;
	}
}

#undef PRINT_ERR
#undef PRINT_WARN
#define PRINT_ERR(X) print_err(cur_file, cur_line, (X))
#define PRINT_WARN(X) print_warn(cur_file, cur_line, (X))

void set_sizes(var *v)
{
	lint *lnt;

	if ((v->ptr || v->fptr) && !v->argarr)
		v->unit_size = 4;
	else if (v->attrs & INT)
	{
		if (v->attrs & SHORT)
			v->unit_size = 2;
		else if (v->attrs & LONG)
			v->unit_size = 8;
		else
			v->unit_size = 4;
	}
	else if (v->attrs & FLOAT)
		v->unit_size = 4;
	else if (v->attrs & DOUBLE)
		v->unit_size = 8;
	else if (v->attrs & CHAR)
		v->unit_size = 1;
	else if (v->attrs & VOID)
		v->unit_size = 0;
	else if (v->attrs & ENUM)
		v->unit_size = 4;
	else if (v->attrs & STRUCT)
		v->unit_size = v->t.s->size;
	else if (v->attrs & UNION)
		v->unit_size = v->t.s->size;

	v->size = v->unit_size;
	if (v->arr_base)
		v->load_size = 4;
	else
		v->load_size = v->unit_size;

	for (lnt = v->arr_base; lnt; lnt = lnt->next)
		v->size *= lnt->n;
}

int is_compound(var *v)
{
	return (v->attrs & (STRUCT | UNION)) || v->arr_base;
}

int parse_declr_name(char *str, var *v, code *parent, char allow_empty)
{
	char name[WORD_LENGTH], *wp = name, sword[WORD_LENGTH];
	char err[WORD_LENGTH], subs[WORD_LENGTH], *sp, pname = 0;
	char width[WORD_LENGTH] = "", *os = str;
	int paren = 0, brack = 0, brace = 0, n;
	var vr, *cur_arg = NULL, *vp;
	char flag, *ost = str;
	lint *lnt, *arr = NULL, *old_arr;

//	printf("PDN: %s\n", str);

	old_arr = v->arr_base;
	v->arr_base = NULL;
	v->bf_size = -1;

	trim_spaces(str);
	if (*str == '(')
	{
		++str;
		pname = 1;
	}
	if (*str == '*')
		for (; *str == '*' || *str == ' '; ++str)
			++v->ptr;

	for (; isalnum(*str) || *str == '_'; ++str)
		*wp++ = *str;
	*wp = '\0';
	if (pname && *str == ')')
		++str;

	wp = sword;
	for (; *str; ++str)
	{
		if (brack)
		{
			if (*str == ']')
			{
				*sp = '\0';
				brack = 0;
				trim_spaces(subs);

				if (subs[0] && !is_any_int(subs))
				{
					PRINT_ERR("Invalid array size subscript");
					return 1;
				}

				lnt = (lint *)calloc(1, sizeof(lint));
				lnt->n = subs[0] ? anytoi(subs) : -1;
				if (v->arr_base)
				{
					lnt->prev = arr;
					if (arr)
						arr->next = lnt;
				}
				else
					v->arr_base = lnt;
				arr = lnt;
			}
			else
				*sp++ = *str;
		}
		else if (paren > 1)
		{
			if (*str == '(' || *str == '[')
				++paren;
			else if (*str == ')' || *str == ']')
				--paren;
			*wp++ = *str;
		}
		else if (paren == 1)
		{
			if (*str == ')' || *str == ',')
			{
				*wp = '\0';
				trim_spaces(sword);

				if (!strcmp(sword, "..."))
				{
					if (v->vararg)
					{
						PRINT_ERR("Parse error in declaration");
						return 1;
					}
					v->vararg = 1;
				}
				else if (!sword[0]);
				else if (!strcmp(sword, "void"));
				else if (parse_declr(sword, &vr, NULL, NULL, 1/*allow_empty*/))
					return 1;
				else if (vr.attrs & (HOST | INPUT | OUTPUT))
				{
					PRINT_ERR("Invalid xCode-specific qualifier in declaration");
					return 1;
				}
				else if (vr.attrs & TYPEDEF)
				{
					PRINT_ERR("Parse error at `typedef'");
					return 1;
				}
				else if ((vr.attrs & EXTERN) || (vr.attrs & STATIC))
				{
					PRINT_ERR("Specified storage class specifier(s) invalid in parameter definition");
					return 1;
				}
				else if (!vr.ptr && ((vr.attrs & STRUCT) ||
					(vr.attrs & UNION)))
				{
					PRINT_ERR("structs/unions may not be passed directly as function parameters");
					return 1;
				}
				else if ((vr.name && !(allow_empty && !vr.name[0]))
					? !is_valid_id(vr.name) : 0)
				{
					sprintf(err, "Invalid identifier `%s'", vr.name);
					PRINT_ERR(err);
					return 1;
				}
				else
				{
					if (v->vararg)
					{
						PRINT_ERR("Parse error in declaration");
						return 1;
					}

					++v->nargs;
					flag = 1;
					for (vp = v->args; vp; vp = vp->next_arg)
						if (vp->name)
							if (vp->name[0] && !strcmp(vr.name, vp->name))
							{
								flag = 0;
								sprintf(err, "Duplicate identifier `%s'", vr.name);
								PRINT_ERR(err);
								return 1;
							}

					if (flag)
					{
						if (vr.arr_base)
							vr.argarr = 1;

						vp = (var *)calloc(1, sizeof(var));
						memcpy(vp, &vr, sizeof(var));

						if (cur_arg)
						{
							cur_arg->next_arg = vp;
							vp->prev_arg = cur_arg;
						}
						else
							v->args = vp;
						cur_arg = vp;
					}

					set_sizes(vp);
				}

				wp = sword;
				if (*str == ')')
					--paren;
			}
			else if (*str == '(' || *str == '[')
			{
				*wp++ = *str;
				++paren;
			}
			else
				*wp++ = *str;
		}
		else if (*str == '(')
		{
			if (paren++)
				*wp++ = *str;
			v->func = 1;
		}
		else if (pname && *str == ')');
		else if (*str == '[')
		{
			sp = subs;
			brack = 1;
		}
		else if (*str == '=')
		{
			*wp = '\0';
			++str;
			trim_spaces(str);
			v->init_val = (char *)calloc(strlen(str)+1, sizeof(char));
			strcpy(v->init_val, str);
			break;
		}
		else if (*str == ':')
		{
			*wp = '\0';
			wp = width;
			for (++str; *str && *str != '='; ++str)
				*wp++ = *str;
			*wp = '\0';
			--str;
		}
		else if (*str != ' ')
		{
			PRINT_ERR("Parse error in declaration");
			return 1;
		}
	}

	if (old_arr)
	{
		if (arr)
			arr->next = old_arr;
		else
			v->arr_base = old_arr;
		old_arr->prev = arr;
	}

	if (pname && v->ptr)
	{
		if (v->func)
		{
			--v->ptr;
			v->func = 0;
			v->fptr = 1;
		}
		else
		{
			v->argarr = 1;
		}
	}

	if (width[0])
	{
		trim_spaces(width);
		n = atoi(width);
		if (n < 0)
			PRINT_ERR("Invalid bitfield width");
		else if (n > 32)
			PRINT_ERR("Maximum bitfield witdh is 32");
		v->bf_size = n;
	}

	trim_spaces(name);
	v->name = (char *)calloc(strlen(name)+1, sizeof(char));
	strcpy(v->name, name);

	return 0;
}

static void pit(char *str, int *sz, int *sz2)
{
	if (*sz + 8 >= *sz2)
		str = realloc(str, *sz2 += WORD_LENGTH);
}

static void piw(char *str, int *sz, int *sz2, char b)
{
	if (*sz + 8 >= *sz2)
		str = realloc(str, *sz2 += WORD_LENGTH);
	str[*sz] = b;
}

int parse_init(char *str, var *v, unsigned char *buf, int *sz, int *sz2)
{
	int j, k, brace = 0, _sz = 0, _sz2 = WORD_LENGTH, ni = 0, needed, i;
	char sword[WORD_LENGTH], *sp = sword, *sp2, first = 0;
	char quote = 0, squote = 0, slash = 0, sconst = 0;
	char *text, *text2, buf2[WORD_LENGTH], flag;
	int stun_size = -1, psz = 0;
	var v2, varg, *smem = NULL, *vp;
	lint *lnt;
	code *acbak;
	_int64 val;

	if (!buf)
	{
		buf = (char *)calloc(WORD_LENGTH, sizeof(char));
		first = 1;
	}
	if (!sz)
	{
		sz = &_sz;
		sz2 = &_sz2;
	}

	memcpy(&v2, v, sizeof(var));

	if (v2.attrs & UNION)
	{
		needed = 1;
		stun_size = v2.t.s->size;
		memcpy(&v2, v2.t.s->var_base, sizeof(var));
	}

	if (lnt = v2.arr_base)
	{
		needed = v2.arr_base->n;
		v2.arr_base = v2.arr_base->next;
	}
	else if (v2.attrs & STRUCT)
	{
		smem = v2.t.s->var_base;
		for (; smem; smem = smem->next)
		{
			if (smem->name[0])
				break;
			i = smem->bf_size;
			if (smem->prev)
				if ((smem->prev->bf_loc + smem->prev->bf_size) & 7)
					i -= 8 - ((smem->prev->bf_loc + smem->prev->bf_size) & 7);
			for (; i > 0; i -= 8)
			{
				if (i >= 8)
				{
					piw(buf, sz, sz2, '\0');
					++*sz;
					++psz;
				}
				else
					piw(buf, sz, sz2, '\0');
			}
		}
	}

	set_sizes(&v2);
	if (str[0] == '{')
	{
		++str;
		if (str[strlen(str)-1] != '}')
			return -1;
		str[strlen(str)-1] = '\0';
	}

	for (; ; ++str)
	{
		if (slash)
		{
			slash = 0;
			*sp++ = *str;
		}
		else if (quote)
		{
			if (*str == '\\')
				slash = 1;
			else if (*str == '"')
				quote = 0;
			*sp++ = *str;
		}
		else if (squote)
		{
			if (*str == '\\')
				slash = 1;
			else if (*str == '\'')
				squote = 0;
			*sp++ = *str;
		}
		else if (*str == '"')
		{
			quote = 1;
			*sp++ = '"';
		}
		else if (*str == '\'')
		{
			squote = 1;
			*sp++ = '\'';
		}
		else if (brace)
		{
			if (*str == '{')
				++brace;
			else if (*str == '}')
				--brace;
			*sp++ = *str;
		}
		else if (*str == '{')
		{
			brace = 1;
			*sp++ = '{';
		}
		else if (*str == ',' || !*str)
		{
			if (*str == ',')
			{
				for (sp2 = str+1; *sp2 == ' '; ++sp2);
				if (!*sp2)
					break;
			}
			++ni;
			*sp = '\0';
			trim_spaces(sword);

			if (v->attrs & STRUCT)
			{
				if (!smem)
				{
					PRINT_ERR("Too many items for `struct' initializer");
					return 1;
				}
				vp = smem;
				while (smem = smem->next)
				{
					if (smem->name[0])
							break;
					i = smem->bf_size;
					if (smem->prev)
						if ((smem->prev->bf_loc + smem->prev->bf_size) & 7)
						{
							++*sz;
							++psz;
							i -= 8 - ((smem->prev->bf_loc + smem->prev->bf_size) & 7);
						}

					flag = 0;
					for (; i > 0; i -= 8)
					{
						if (i >= 8)
						{
							piw(buf, sz, sz2, '\0');
							++*sz;
							++psz;
							flag = 1;
						}
						else if (flag)
							piw(buf, sz, sz2, '\0');
					}
				}
			}
			else
				vp = &v2;

			if (is_string(sword))
			{
				if (!vp->ptr && (vp->attrs & CHAR) && v->arr_base)
				{
					sconst = 1;
					sword[strlen(sword)-1] = '\0';
					text = (char *)calloc(strlen(sword), sizeof(char));
					j = translate_string(text, sword+1);
					if (v->arr_base->n != -1 && j > v->arr_base->n)
					{
						pit(buf, sz, sz2);
						memcpy(&buf[*sz], text, v->arr_base->n);
						*sz += v->arr_base->n;
						psz += v->arr_base->n;
					}
					else
					{
						pit(buf, sz, sz2);
						memcpy(&buf[*sz], text, j);
						*sz += j;
						psz += j;
					}
					if (j < v->arr_base->n)
					{
						pit(buf, sz, sz2);
						memset(&buf[*sz], 0, v->arr_base->n - j);
						*sz += v->arr_base->n - j;
						psz += v->arr_base->n - j;
					}
					free(text);
				}
				else if (vp->ptr && (vp->attrs & CHAR))
				{
					text = (char *)calloc(strlen(sword), sizeof(char));
					sword[strlen(sword)-1] = '\0';
					j = translate_string(text, sword+1);
					text2 = (char *)calloc(9 + j * 4, sizeof(char));

					sprintf(text2, ".db %u", text[0]);
					for (k = 1; k < j; ++k)
					{
						sprintf(buf2, ",%u", text[k]);
						strcat(text2, buf2);
					}
					acbak = asm_code;
					asm_code = global.code_base;
					write_asm(text2);

					pit(buf, sz, sz2);
					pack_int2(&buf[*sz], global.code_base->self_size,
						4);
					*sz += 4;
					psz += 4;

					sprintf(buf2, "ld a,%d", asm_code->self_size);
					write_asm(buf2);
					sprintf(buf2, "ld b,%d", cur_data++);
					write_asm(buf2);
					write_asm("dcpy a,b");
					asm_code->self_size += j;
					asm_code = acbak;

					free(text);
					free(text2);
				}
			}
			else
			{
				trim_spaces(sword);
				if (sword[0] == '{')
				{
					if (parse_init(sword, vp, buf, sz, sz2) == -1)
						return -1;
				}
				else
				{
					errno = 0;
					val = eval(sword, &varg, 1, 0);
					if (errno)
						return 1;
					if (!vp->arr_base)
					{
						if (vp->bf_size != -1)
						{
							if ((vp->attrs & SIGNED) && val < 0)
								val += (_int64)1 << vp->bf_size;

							if (vp->bf_loc)
							{
								flag = 0;
								buf[*sz] |= (val & ((1 << vp->bf_size)-1)) << ((vp->prev->bf_loc + vp->prev->bf_size) & 7);
								i = vp->bf_size - (9 - vp->bf_loc - 1);
								val >>= 8 - ((vp->prev->bf_loc + vp->prev->bf_size) & 7);
								if (i <= 0)
								{
									if (!i)
									{
										++*sz;
										++psz;
									}
								}
								else for (++*sz, ++psz; i > 0; i -= 8)
								{
									if (i >= 8)
									{
										piw(buf, sz, sz2, val & 255);
										++*sz;
										if (i == 8)
										{
											++*sz;
											psz += 2;
										}
										else
										{
											++psz;
											val >>= 8;
										}
									}
									else
										piw(buf, sz, sz2, val & ((1 << i) - 1));
								}
							}
							else
							{
								for (i = vp->bf_size; i > 0; i -= 8)
								{
									if (i >= 8)
									{
										piw(buf, sz, sz2, val & 255);
										++*sz;
										++psz;
										val >>= 8;
									}
									else
										piw(buf, sz, sz2, val & ((1 << i) - 1));
								}
							}
						}
						else if (vp->attrs & (INT | CHAR | ENUM))
						{
							if (vp->prev)
								if(vp->prev->bf_size > 0
									&& ((vp->prev->bf_loc + vp->prev->bf_size) & 7))
								{
									++*sz;
									++psz;
								}
							pit(buf, sz, sz2);
							if (varg.attrs & (FLOAT | DOUBLE))
								pack_int2(&buf[*sz], *(_float64 *)&val, vp->unit_size);
							else
								pack_int2(&buf[*sz], val, vp->unit_size);
							*sz += vp->unit_size;
							psz += vp->unit_size;
						}
						else if (vp->attrs & (FLOAT | DOUBLE))
						{
							if (vp->prev)
								if(vp->prev->bf_size > 0
									&& ((vp->prev->bf_loc + vp->prev->bf_size) & 7))
								{
									++*sz;
									++psz;
								}
							if (varg.attrs & (FLOAT |DOUBLE))
								sprintf(sword, "%.20f", val);
							else
								sprintf(sword, "%d.0", (int)val);
							pit(buf, sz, sz2);
							pack_float(&buf[*sz], sword, get_sn(vp->unit_size));
							*sz += vp->unit_size;
							psz += vp->unit_size;
						}
					}
				}
			}
			sp = sword;
		}
		else
			*sp++ = *str;

		if (!*str)
			break;
	}

	if (v->arr_base)
	{
		if (needed == -1)
		{
			if (sconst)
				v->arr_base->n = lnt->n = *sz;
			else
				v->arr_base->n = lnt->n = ni;
		}
		else if (ni < needed)
		{
			needed -= ni;
			for (j = 0; j < needed; ++j)
				for (k = 0; k < v2.unit_size; ++k)
				{
					piw(buf, sz, sz2, 0);
					++*sz;
				}
		}
		else if (ni > needed)
		{
			PRINT_ERR("Too many items in initializer");
			return -1;
		}
	}
	else if (v->attrs & STRUCT)
	{
		if (vp)
			if(vp->bf_size > 0 && (vp->bf_loc + vp->bf_size) & 7)
			{
				++*sz;
				++psz;
			}
		pit(buf, sz, sz2);
		for (; psz <= v->t.s->size; ++psz)
		{
			piw(buf, sz, sz2, 0);
			++*sz;
		}
	}
	else if (stun_size != -1 && psz < stun_size)
	{
		if (v2.bf_size != -1)
			if ((v2.bf_loc + v2.bf_size) & 7)
			{
				++*sz;
				++psz;
			}
		pit(buf, sz, sz2);
		i = stun_size - psz;
		*sz += i;
		pit(buf, sz, sz2);
		memset(&buf[*sz-i], 0, stun_size - psz);
	}

	if (first)
	{
		if (!*sz)
		{
			PRINT_ERR("Empty initializer");
			return -1;
		}
		free(calloc(99999, sizeof(char)));
		text = (char *)calloc(7 + *sz * 4, sizeof(char));
		sprintf(text, ".db %u", (unsigned int)buf[0]);
		for (j = 1; j < *sz; ++j)
		{
			sprintf(buf2, ",%u", (unsigned int)buf[j]);
			strcat(text, buf2);
		}
		acbak = asm_code;
		asm_code = global.code_base;
		write_asm(text);
		asm_code = acbak;
		free(text);

		if (v->global)
		{
			if (v->attrs & STATIC)
			{
				acbak = asm_code;
				asm_code = global.code_base;
			}
			sprintf(buf2, "ld a,%d",
				v2.loc);
			write_asm(buf2);
		}
		else
		{
			sprintf(buf2, "laddr %d",
				v2.loc);
			write_asm(buf2);
			write_asm("pop a");
		}

		sprintf(buf2, "ld b,%d", cur_data++);
		write_asm(buf2);
		write_asm("dcpy a,b");

		if (v->global && (v->attrs & STATIC))
			asm_code = acbak;

		return 1;
	}

	return 0;
}

int parse_declr(char *str, var *v, code *parent, stun *sup, char allow_empty)
{
	char sword[WORD_LENGTH], err[WORD_LENGTH], *sp = sword;
	char mode = 0, flag, heres_one = 0, did_one = 0, need_tag = 0;
	char flag2;
	char buf[WORD_LENGTH], buf2[WORD_LENGTH], tname[WORD_LENGTH] = "", *sp2;
	unsigned int j, atrs, en = 0;
	int paren = 0, brace = 0;
	var vr, *vp, *vp2, varg;
	code *c, *acbak, *add_here;
	func *f;
	lint *lnt, *lnt2, *cur_lint;
	lstr *ls;
	enumt *e;
	econst *ec;
	stun *su;
	char *os = str;
	memset(v, 0, sizeof(var));

	v->ptr = v->func = 0;
	v->arr_base = NULL;
	v->args = NULL;
	for (; ; ++str)
	{
		if (mode)
		{
			if (paren)
			{
				if (*str == '(' || *str == '[' || *str == '{')
					++paren;
				else if (*str == ')' || *str == ']' || *str == '}') 
					--paren;
				*sp++ = *str;
			}
			else if (*str == '(' || *str == '[' || *str == '{')
			{
				paren = 1;
				*sp++ = *str;
			}
			else if (!*str || *str == ',')
			{
				*sp = '\0';
A_Var:			if (!asm_code)
				{
					memcpy(&vr, v, sizeof(var));
					if (parse_declr_name(sword, &vr, parent ? parent : global.code_base, !!sup || allow_empty))
						return 1;

					if (vr.name[0] && (vr.attrs & TYPEDEF))
					{
						ls = (lstr *)calloc(1, sizeof(lstr));
						ls->str = (char *)calloc(strlen(vr.name)+1,
							sizeof(char));
						strcpy(ls->str, vr.name);
						ls->next = temp_td_base;
						if (temp_td_base)
							temp_td_base->prev = ls;
						temp_td_base = ls;
					}
				}
				if (is_cast_chars(sword))
				{
					if (v->attrs & TYPEDEF)
					{
						PRINT_ERR("Erroneous or no type name specified");
						return 1;
					}
					if (parse_declr_name(sword, v, NULL, 1))
						return 1;
					set_sizes(v);
					return 0;
				}
				else
				{
					memcpy(&vr, v, sizeof(var));
					if (parse_declr_name(sword, &vr, parent ? parent : global.code_base, !!sup || allow_empty))
					{
						if (vr.attrs & TYPEDEF)
							PRINT_ERR("Parse error");
						return 1;
					}
				}
				did_one = 1;

				if (!is_valid_id(vr.name) && !((sup || allow_empty) && !vr.name[0]))
				{
					sprintf(err, "Invalid identifier `%s'",
						vr.name);
					PRINT_ERR(err);
					print_var_info_stdout(&vr);
					puts(os);
					return 1;
				}

				if (parent && !sup)
				if (!id_free(vr.name, parent, NULL))
				{
					sprintf(err, "Duplicate identifier `%s'",
						vr.name);
					PRINT_ERR(err);
					return 1;
				}

				set_sizes(&vr);
				if (vr.attrs & TYPEDEF)
				{
					if (vr.init_val)
					{
						PRINT_ERR("Initializer specified with `typedef'");
						return 1;
					}

					vr.attrs ^= TYPEDEF;
					vp = (var *)malloc(sizeof(var));
					memcpy(vp, &vr, sizeof(var));
					if (parent)
					{
						vp->next = parent->td_base;
						if (parent->td_base)
							parent->td_base->prev = vp;
						parent->td_base = vp;
					}
					else if (asm_code)
					{
						vp->next = global.code_base->td_base;
						if (global.code_base->td_base)
							global.code_base->td_base->prev = vp;
						global.code_base->td_base = vp;
					}
				}
				else if (parent)
				{
					if (sup)
					{
						if (vr.attrs & XCATTRS)
						{
							PRINT_ERR("Invalid xCode specific qualifier in declaration");
							return 1;
						}

						for (vp = sup->var_base; vp; vp = vp->next)
							if (vp->name[0])
								if (!strcmp(vr.name, vp->name))
								{
									sprintf(err, "Duplicate identifier `%s'\n",
										vr.name);
									PRINT_ERR(err);
									return 1;
								}

						vp = (var *)malloc(sizeof(var));
						memcpy(vp, &vr, sizeof(var));

						if (sup->var_base)
						{
							sup->cur_var->next = vp;
							vp->prev = sup->cur_var;
						}
						else
							sup->var_base = vp;
						sup->cur_var = vp;
					}
					else if (vr.func == 1)
					{
						for (f = func_base; f; f = f->next)
							if (!strcmp(f->type.name, vr.name))
							{
								if (typecmp(&f->type, &vr))
								{
									PRINT_ERR("Conflicting function definitions/prototypes");
									return 1;
								}
								break;
							}

						if (!f)
						{
							if (vr.attrs & HOST)
							{
								if (settings.host_table)
									printf("[%d] %s\n", host_funcs, vr.name);
								vr.host = host_funcs++;
							}
							f = (func *)calloc(1, sizeof(func));
							memcpy(&f->type, &vr, sizeof(var));
							f->next = func_base;
							if (func_base)
								func_base->prev = f;
							func_base = f;
						}
					}
					else
					{
						if (vr.attrs & XCATTRS)
						{
							PRINT_ERR("Invalid xCode specific qualifier in declaration");
							return 1;
						}

						vp = (var *)malloc(sizeof(var));
						memcpy(vp, &vr, sizeof(var));
						vp->next = parent->var_base;
						if (parent->var_base)
							parent->var_base->prev = vp;
						parent->var_base = vp;

						if (vp->attrs & STATIC)
							add_here = global.code_base;
						else
							add_here = parent;
						vp->loc = add_here->size;
						vp->global = (parent == global.code_base) ||
							(vr.attrs & STATIC);

						if (vp->init_val)
						{
							if (is_compound(vp))
							{
								if (is_valid_id(vp->init_val))
								{
									if (!(vp2 = find_var(vp->init_val, asm_code)))
									{
										PRINT_ERR("Invalid initializer");
										return 1;
									}
									if (typecmp(vp, vp2))
									{
										PRINT_ERR("Incompatible type variable given as initializer");
										return 1;
									}

									if (vp->global)
									{
										sprintf(buf, "ld a,%d", vp->loc);
										write_asm(buf);
									}
									else
									{
										sprintf(buf, "laddr %d", vp->loc);
										write_asm(buf);
										write_asm("pop a");
									}
									if (vp2->global)
									{
										sprintf(buf, "ld b,%d", vp2->loc);
										write_asm(buf);
									}
									else
									{
										sprintf(buf, "laddr %d", vp2->loc);
										write_asm(buf);
										write_asm("pop b");
									}

									sprintf(buf, "ld c,%d", vp->size);
									write_asm(buf);
									write_asm("cpy a,b,c");
								}
								else
								{
									parse_init(vp->init_val, vp,
										NULL, NULL, NULL);
									set_sizes(vp);
								}
							}
							else if (vp->global)
							{
								trim_spaces(vp->init_val);
								if (vp->init_val[0] == '{' &&
									vp->init_val[strlen(vp->init_val)-1] == '}')
								{
									vp->init_val[strlen(vp->init_val)-1] = '\0';
									++vp->init_val;
								}

								if (parent != global.code_base)
								{
									acbak = asm_code;
									asm_code = global.code_base;
								}
								eval(vp->init_val, &varg, 0, 0);

								if (typecmp(&varg, vp))
									cast(&varg, vp);
								sprintf(buf, "pop%d b",
									varg.load_size);
								write_asm(buf);
								sprintf(buf, "ld%d (%d),b",
									varg.load_size, vp->loc);
								write_asm(buf);
								if (parent != global.code_base)
									asm_code = acbak;
							}
							else
							{
								eval(vp->init_val, &varg, 0, 0);
								if (typecmp(&varg, vp))
									cast(&varg, vp);
								sprintf(buf, "laddr %d", vp->loc);
								write_asm(buf);
								write_asm("pop a");
								sprintf(buf, "pop%d b",
									varg.load_size);
								write_asm(buf);
								sprintf(buf, "ld%d (a),b",
									varg.load_size);
								write_asm(buf);
							}
						}
						else if (vp->attrs & STATIC)
						{
							if (parent != global.code_base)
							{
								acbak = asm_code;
								asm_code = global.code_base;
							}
							sprintf(buf, "ld a,%d", vp->loc);
							write_asm(buf);
							sprintf(buf, "ld b,%d", vp->size);
							write_asm(buf);
							write_asm("zero a,b");
							if (parent != global.code_base)
								asm_code = acbak;
						}

						add_here->size = (add_here->self_size +=
							vp->size) + add_here->child_size;
						for (c = add_here->parent; c &&
							c != global.code_base; c = c->parent)
						{
							if (c->parent->size > c->child_size)
							{
								c->size = c->self_size +
									(c->child_size =
									c->parent->size);
							}
							else
								break;
						}
					}
				}
				else
					memcpy(v, &vr, sizeof(var));
				sp = sword;
			}
			else
				*sp++ = *str;
		}
		else
		{
			if (paren)
			{
				if (*str == '(' || *str == '[' || *str == '{')
					++paren;
				else if (*str == ')' || *str == ']' || *str == '}')
					--paren;
			}
			else if (*str == '(' || *str == '[' || *str == '{')
				paren = 1;
			if (!*str || *str == ' ' || *str == '*' || *str == '(' ||
				*str == '[' || *str == '{' || *str == ',' || *str == '=')
			{
				*sp = '\0';
				flag = 1;

				if (need_tag == 1 && *str == '{')
					need_tag = 2;

				if (need_tag == 1)
				{
					flag = 0;
					strcpy(tname, sword);

					if (v->attrs & ENUM)
					{
						if (e = find_enum(tname, asm_code))
							v->t.e = e;
					}
					else if (v->attrs & STRUCT)
					{
						if (su = find_struct(tname, asm_code))
							v->t.s = su;
						else
						{
							su = (stun *)calloc(1, sizeof(stun));
							v->t.s = su;
							if (tname[0])
							{
								su->name = (char *)calloc(strlen(tname)+1,
									sizeof(char));
								strcpy(su->name, tname);
							}
							su->size = -1;
						}	
					}
					else if (v->attrs & UNION)
					{
						if (su = find_union(tname, asm_code))
							v->t.s = su;
						else
						{
							su = (stun *)calloc(1, sizeof(stun));
							v->t.s = su;
							if (tname[0])
							{
								su->name = (char *)calloc(strlen(tname)+1,
									sizeof(char));
								strcpy(su->name, tname);
							}
							su->size = -1;
						}	
					}

					need_tag = 2;
				}
				else if (need_tag == 2 && *str == '{')
				{
					flag = 0;
					need_tag = 0;
					--paren;
					buf2[0] = '\0';
					if (v->t.e)
						if (v->attrs & ENUM)
						{
							PRINT_ERR("Redefinition of tag");
							return 1;
						}
						else
						{
							if (v->t.s->size > -1)
							{
								PRINT_ERR("Redefinition of tag");
								return 1;
							}
						}
					if (v->attrs & ENUM)
					{
						e = (enumt *)calloc(1, sizeof(enumt));
						v->t.e = e;
						if (tname[0])
						{
							e->name = (char *)calloc(strlen(tname)+1,
								sizeof(char));
							strcpy(e->name, tname);
						}

						sp2 = buf;
						for (++str; *str; ++str)
						{
							if (*str == ',' || *str == '}')
							{
								*sp2 = '\0';
								trim_spaces(buf);
								trim_spaces(buf2);

								if ((buf2[0] && is_any_int(buf2))
									|| !buf2[0])
								{
									if (buf2[0])
										en = anytoi(buf2);

									if (!is_valid_id(buf))
									{
										sprintf(err, "Invalid identifier `%s'", buf);
										PRINT_ERR(err);
									}
									else if (!id_free(buf, parent, NULL))
									{
										sprintf(err, "Duplicate identifier `%s'", buf);
										PRINT_ERR(err);
									}
									else
									{
										ec = (econst *)calloc(1, sizeof(econst));
										ec->name = (char *)calloc(strlen(buf)+1, sizeof(char));
										strcpy(ec->name, buf);
										ec->val = en++;
										ec->next = e->econst_base;
										e->econst_base = ec;
									}
								}
								else
								{
									sprintf(buf, "Illegal value `%s' specified for `enum' constant", buf);
									PRINT_ERR(buf);
								}
								if (*str == '}')
									break;
								sp2 = buf;
								buf2[0] = '\0';
							}
							else if (*str == '=')
							{
								*sp2 = '\0';
								sp2 = buf2;
							}
							else
								*sp2++ = *str;
						}

						if (parent)
						{
							e->next = parent->enum_base;
							parent->enum_base = e;
						}
					}
					else if (v->attrs & STRUCT)
					{
						if (!v->t.s)
							v->t.s = (stun *)calloc(1, sizeof(stun));
						su = v->t.s;
						su->size = 0;

						sp2 = buf;
						for (++str; *str; ++str)
						{
							if (paren)
							{
								if (*str == '(' || *str == '['
									|| *str == '{')
									++paren;
								else if (*str == ')' ||
									*str == ']' || *str == '}')
									--paren;
								*sp2++ = *str;
							}
							else if (*str == '(' || *str == '['
								|| *str == '{')
							{
								++paren;
								*sp2++ = *str;
							}
							else if (*str == '}')
								break;
							else if (*str == ';')
							{
								*sp2 = '\0';
								trim_spaces(buf);
								parse_declr(buf, &varg, asm_code, su, 0);
								sp2 = buf;
							}
							else
								*sp2++ = *str;
						}

						if (parent)
						{
							su->next = parent->struct_base;
							parent->struct_base = su;
						}

						for (vp = su->var_base; vp; vp = vp->next)
						{
							if (vp->bf_size != -1)
							{
								flag2 = 1;
								if (vp->prev)
									if (vp->prev->bf_size > 0
										&& (vp->prev->bf_loc + vp->prev->bf_size) & 7)
									{
										flag2 = 0;
										vp->loc = su->size;
										vp->bf_loc = (vp->prev->bf_loc + vp->prev->bf_size) & 7;
										su->size += (vp->bf_loc + vp->bf_size) >> 3;
									}
								if (flag2)
								{
									vp->loc = su->size;
									vp->bf_loc = 0;
									su->size += vp->bf_size >> 3;
								}
							}
							else
							{
								if (vp->prev)
									if (vp->prev->bf_size > 0
										&& (vp->prev->bf_loc + vp->prev->bf_size) & 7)
										++su->size;
								vp->loc = su->size;
								su->size += vp->size;
							}
						}
					}
					else if (v->attrs & UNION)
					{
						if (!v->t.s)
							v->t.s = (stun *)calloc(1, sizeof(stun));
						su = v->t.s;
						su->size = 0;

						sp2 = buf;
						for (++str; *str; ++str)
						{
							if (paren)
							{
								if (*str == '(' || *str == '['
									|| *str == '{')
									++paren;
								else if (*str == ')' ||
									*str == ']' || *str == '}')
									--paren;
								*sp2++ = *str;
							}
							else if (*str == '(' || *str == '['
								|| *str == '{')
							{
								++paren;
								*sp2++ = *str;
							}
							else if (*str == '}')
								break;
							else if (*str == ';')
							{
								*sp2 = '\0';
								trim_spaces(buf);
								parse_declr(buf, &varg, asm_code, su, 0);
								sp2 = buf;
							}
							else
								*sp2++ = *str;
						}

						if (parent)
						{
							su->next = parent->union_base;
							parent->union_base = su;
						}

						for (vp = su->var_base; vp; vp = vp->next)
						{
							if (vp->bf_size != -1)
							{
								if (vp->bf_size & 7)
								{
									if ((vp->bf_size >> 3) + 1 > su->size)
										su->size = (vp->bf_size >> 3) + 1;
								}
								else
								{
									if ((vp->bf_size >> 3) > su->size)
										su->size = (vp->bf_size >> 3);
								}
							}
							else if (vp->size > su->size)
								su->size = vp->size;
						}
					}
				}
				else if (vp = find_td(sword, asm_code))
				{
					flag = 0;
					for (j = 0; j < N_ATTRS; ++j)
						if (vp->attrs & (1 << j))
							v->attrs |= (1 << j);
					v->ptr += vp->ptr;
					v->t.e = vp->t.e;
					if (vp->arr_base)
						if (v->arr_base)
						{
							PRINT_ERR("Conflicting type attributes due to `typedef'");
							return 1;
						}
						else
						{
							cur_lint = NULL;
							for (lnt = vp->arr_base; lnt; lnt = lnt->next)
							{
								lnt2 = (lint *)calloc(1, sizeof(lint));
								lnt2->n = lnt->n;
								if (cur_lint)
								{
									cur_lint->next = lnt2;
									lnt2->prev = cur_lint;
								}
								else
									v->arr_base = lnt2;
								cur_lint = lnt2;
							}
						}
				}
				else for (j = 0; j < N_ATTRS; ++j)
					if (!strcmp(sword, attrs[j]))
					{
						flag = 0;
						if (v->attrs & (1 << j))
						{
							sprintf(err, "`%s' repeated in declaration",
								sword);
							PRINT_ERR(err);
							return 1;
						}
						else
						{
							v->attrs |= (1 << j);
							if (NEEDS_ARG(j))
								need_tag = 1;
						}
						break;
					}

				if (flag)
				{
					if (!*str || *str == ',')
					{
						heres_one = 1;
						sp = sword;
					}
					else
						*sp++ = *str;
					mode = 1;
				}
				else if (*str == ' ' || *str == '}')
				{
					sp = sword;
				}
				else
				{
					*sword = *str;
					sp = sword+1;
					mode = 1;
				}

				if (mode)
				{
					atrs = v->attrs;

					if ((atrs & (ENUM | UNION | STRUCT)) && 
						!v->t.e)
					{
						PRINT_ERR("Tag used but not formally defined");
						return 1;
					}

					if (settings.ansi_on && (atrs & XCATTRS))
					{
						PRINT_ERR("xcc-specific specifier with -ansi and not in #pragma unansi");
						return 1;
					}

					if ((atrs & SIGNED) && (atrs & UNSIGNED))
					{
						PRINT_ERR("Declaration specifies both `signed' and `unsigned'");
						return 1;
					}
					if ((atrs & SHORT) && (atrs & LONG))
					{
						PRINT_ERR("Declaration specifies both `short' and `long'");
						return 1;
					}

					if (atrs & LONG)
					{
						if ((atrs & VOID) || (atrs & CHAR) ||
							(atrs & FLOAT) || (atrs & STRUCT) ||
							(atrs & UNION) || (atrs & ENUM))
						{
							PRINT_ERR("`long' incompatible with declaration type");
							return 1;
						}

						if (!(atrs & DOUBLE))
							v->attrs = (atrs |= INT);
					}

					if (atrs & SHORT)
					{
						if ((atrs & VOID) || (atrs & CHAR) ||
							(atrs & FLOAT) || (atrs & STRUCT) ||
							(atrs & UNION) || (atrs & ENUM) ||
							(atrs & DOUBLE))
						{
							PRINT_ERR("`short' incompatible with declaration type");
							return 1;
						}

						v->attrs = (atrs |= INT);
					}

					if ((atrs & SIGNED) || (atrs & UNSIGNED))
					{
						if ((atrs & VOID) || (atrs & FLOAT) ||
							(atrs & STRUCT) || (atrs & UNION) ||
							(atrs & ENUM) || (atrs & DOUBLE))
						{
							PRINT_ERR("'signed'/'unsigned' incompatible with declaration type");
							return 1;
						}

						if (!(atrs & CHAR))
							v->attrs = (atrs |= INT);
					}

					flag = 0;
					if (atrs & INT)
						++flag;
					if (atrs & FLOAT)
						++flag;
					if (atrs & DOUBLE)
						++flag;
					if (atrs & VOID)
						++flag;
					if (atrs & CHAR)
						++flag;
					if (atrs & STRUCT)
						++flag;
					if (atrs & ENUM)
						++flag;
					if (atrs & UNION)
						++flag;

					if (flag > 1)
					{
						PRINT_ERR("Conflicting types in declaration");
						print_var_info_stdout(v);
						return 1;
					}
					if (!flag)
						v->attrs = (atrs |= INT);

					if (((atrs & INT) || (atrs & CHAR)) &&
						!(atrs & SIGNED) && !(atrs & UNSIGNED))
						v->attrs = (atrs |= SIGNED);

					flag = 0;
					if (atrs & TYPEDEF)
						++flag;
					if (atrs & EXTERN)
						++flag;
					if (atrs & STATIC)
						++flag;
					if (atrs & AUTO)
						++flag;
					if (atrs & REGISTER)
						++flag;

					if (flag > 1)
					{
						if (atrs & TYPEDEF)
							PRINT_ERR("Storage class specified in typedef");
						else
							PRINT_ERR("Multiple storage classes specified in single declaration");
						return 1;
					}

					if (!flag)
						v->attrs = (atrs |= AUTO);

					if (heres_one)
					{
						heres_one = 0;
						goto A_Var;
					}
				}
			}
			else
				*sp++ = *str;
		}
		if (!*str)
			break;
	}

	if (!did_one)
		set_sizes(v);
	return 0;
}

char *def_label;
caset *case_base;
code *swbase;

void switch_search(code *c)
{
	char buf[WORD_LENGTH];
	lstr *ls, *ls2;
	caset *cs;
	var v;
	code *c2;
	int n;

	if (c->type == MULTI)
	{
		for (c2 = c->child; c2; c2 = c2->next)
			switch_search(c2);
	}

	for (ls = c->label_base; ls; ls = ls2)
	{
		ls2 = ls->next;
		if (!strcmp(ls->str, "default"))
		{
			if (def_label)
				PRINT_ERR("Multiple `default' labels in single `switch' block");
			else
			{
				def_label = new_label();

				if (ls->prev)
					ls->prev->next = ls->next;
				else
					c->label_base = ls->next;
				if (ls->next)
					ls->next->prev = ls->prev;

				ls->str = def_label;
				ls->next = c->forced_label_base;
				if (c->forced_label_base)
					c->forced_label_base->prev = ls;
				c->forced_label_base = ls;
			}
		}
		else if (!strcmp(first_word(buf, ls->str), "case"))
		{
			n = eval(&ls->str[4], &v, 1, 0);

			for (cs = case_base; cs; cs = cs->next)
				if (n == cs->val)
				{
					print_err(cur_file, c->pos,
						"Duplicate `case' value");
					return;
				}

			cs = (caset *)calloc(1, sizeof(caset));
			cs->next = case_base;
			if (case_base)
				case_base->prev = cs;
			case_base = cs;

			case_base->val = n;
			case_base->label = new_label();

			if (ls->prev)
				ls->prev->next = ls->next;
			else
				c->label_base = ls->next;
			if (ls->next)
				ls->next->prev = ls->prev;

			ls->str = case_base->label;
			ls->next = c->forced_label_base;
			if (c->forced_label_base)
				c->forced_label_base->prev = ls;
			c->forced_label_base = ls;
		}
	}

	if (c->child)
	{
		if (c->str ? strcmp(first_word(buf, c->str), "switch") : 1)
			switch_search(c->child);
	}
	if (c->peer)
		switch_search(c->peer);
}

code *code_end(code *c)
{
	if (c->peer)
	{
		if (c->peer->child)
			return c->peer->child;
		else
			return c->peer;
	}
	else if (c->child)
		return c->child;
	else
		return c;
}

void do_statements(code *c, code *parent, char *_break,
				   char *_continue)
{
	code *tc, *c2;
	var v, *vp;
	func *f;
	char sword[WORD_LENGTH], *sp = sword, *cp, buf[WORD_LENGTH];
	char flag, *sp2, *sp3;
	lstr *ls, *ls2;
	ulbl *ul;
	caset *cs, *cs2;
	pevent *pe;
	char paren = 0, quote = 0, squote = 0, slash = 0;

	cur_file = c->file;
	cur_line = c->pos;

	while (pevent_base)
	{
		if (cur_line < pevent_base->pos)
			break;
		if (!strcmp(pevent_base->str, "unansi"))
			settings.ansi_on = 0;
		else if (!strcmp(pevent_base->str, "reansi"))
			settings.ansi_on = settings.ansi;
		else if (!strncmp(pevent_base->str, "hostfunc", 8))
		{
			strtok(pevent_base->str, " ");
			if (sp2 = strtok(NULL, " "))
				host_funcs = atoi(sp2);
		}
		pe = pevent_base->next;
		free(pevent_base);
		pevent_base = pe;
	}

	for (ls = c->forced_label_base; ls; ls = ls2)
	{
		ls2 = ls->next;
		sprintf(buf, "%s:", ls->str);
		write_asm(buf);
		free(ls);
	}

	for (ls = c->label_base; ls; ls = ls->next)
	{
		if (strcmp(ls->str, "default") && strncmp(ls->str, "case ", 5))
		{
			if (asm_code == global.code_base)
			{
				PRINT_ERR("Label found not within a function");
				return;
			}
			if (!is_valid_id(ls->str))
			{
				sprintf(buf, "Invalid identifier `%s'", ls->str);
				PRINT_ERR(buf);
				return;
			}
			if (!id_free(ls->str, asm_code, func_base))
			{
				sprintf(buf, "Duplicate identifier `%s'", ls->str);
				PRINT_ERR(buf);
				return;
			}

			for (ul = func_base->orphan_base; ul; ul = ul->next)
				if (!strcmp(ul->user, ls->str))
					break;

			if (ul)
			{
				if (ul->prev)
					ul->prev->next = ul->next;
				else
					func_base->orphan_base = ul->next;
				if (ul->next)
					ul->next->prev = ul->prev;
			}
			else
			{
				ul = (ulbl *)calloc(1, sizeof(ulbl));

				ul->user = (char *)calloc(strlen(ls->str)+1,
					sizeof(char));
				strcpy(ul->user, ls->str);
				ul->real = new_label();
			}
			
			ul->next = func_base->label_base;
			if (func_base->label_base)
				func_base->label_base->prev = ul;
			func_base->label_base = ul;

			sprintf(buf, "%s:", ul->real);
			write_asm(buf);
		}
	}

	switch (c->type)
	{
	case MULTI:
		if (c->prev)
			if (c->prev->type == FUNC)
				asm_code = c;

		for (tc = c->child; tc; tc = tc->next)
			do_statements(tc, c, c->_break ? c->_break : _break,
			c->_continue ? c->_continue : _continue);
		if (asm_code == c)
		{
			if (asm_code != global.code_base)
			for (ul = func_base->orphan_base; ul; ul = ul->next)
			{
				sprintf(buf, "Reference to undefined label `%s'",
					ul->user);
				print_err(cur_file, ul->pos, buf);
			}

			asm_code = global.code_base;
		}

		break;
	case FUNC:
		if (parse_declr(c->str, &v, NULL, NULL, 0));
		else if (v.attrs & TYPEDEF)
			PRINT_ERR("Parse error at `typedef'");
		else if ((v.attrs & EXTERN) || (v.attrs & REGISTER) ||
			(v.attrs & CONST) || (v.attrs & VOLATILE))
			PRINT_ERR("Given storage class and type qualifier(s) invalid in function declaration");
		else if (!v.ptr && ((v.attrs & STRUCT) || (v.attrs & UNION)))
			PRINT_ERR("Function may not return a struct or union directly");
		else
		{
			for (f = func_base; f; f = f->next)
				if (!strcmp(f->type.name, v.name))
				{
					if (f->code_base)
					{
						sprintf(buf, "Redefinition of function `%s'",
							v.name);
						PRINT_ERR(buf);
						return;
					}
					else if (typecmp(&f->type, &v))
					{
						PRINT_ERR("Conflicting function definitions/prototypes");
						return;
					}

					free_var(&f->type, 0);
					memcpy(&f->type, &v, sizeof(var));
					f->code_base = c->next;
					break;
				}

			if (!f)
			{
				for (vp = v.args; vp; vp = vp->next_arg)
					if (!vp->name[0])
					{
						PRINT_ERR("Name not specified for function argument");
						f = (func *)1;
						break;
					}
			}

			if (!f)
			{
				f = (func *)calloc(1, sizeof(func));
				memcpy(&f->type, &v, sizeof(var));
				f->code_base = c->next;
				f->next = func_base;
				if (func_base)
					func_base->prev = f;
				func_base = f;
			}

			for (vp = f->type.args; vp; vp = vp->next)
			{
				vp->next = vp->next_arg;
				vp->prev = vp->prev_arg;
				vp->loc = f->code_base->size;
				if (vp->argarr)
					f->code_base->size += 4;
				else
					f->code_base->size += vp->size;
			}
			f->code_base->self_size =
				f->code_base->size;
			f->code_base->var_base = f->type.args;
		}
		break;
	case DECLR:
//		printf("Declr: %s\n", c->str);
		parse_declr(c->str, &v, parent, NULL, 1);
		break;
	case CMND:
		if (parent == global.code_base)
		{
			PRINT_ERR("Parse error");
			return;
		}

		sprintf(buf, "; Statement: %s", c->str);
		write_asm(buf);
//		printf("Statement: %s\n", c->str);

		for (cp = c->str; *cp; ++cp)
			if (!isalpha(*cp) && !isdigit(*cp))
				break;
			else
				*sp++ = *cp;
		*sp = '\0';

		if (!strcmp(sword, "if"))
		{
			eval(cp, &v, 0, 0);
			sprintf(buf, "pop%d a", v.load_size);
			write_asm(buf);
			sprintf(buf, "test%d a", v.load_size);
			write_asm(buf);
			if (!c->child)
			{
				PRINT_ERR("No statement following `if'");
				return;
			}

			flag = 0;
			if (c->peer)
				if (!strcmp(c->peer->str, "else"))
					flag = 1;

			if (flag)
			{
				if (!c->peer->child)
					return;
				sprintf(buf, "jp z,%s", sp = new_label());
				write_asm(buf);
				sprintf(buf, "jp %s", sp2 = new_label());
				queue_asm(c->child, buf);
				sprintf(buf, "%s:", sp2);
				queue_asm(c->peer->child, buf);
				sprintf(buf, "%s:", sp);
				queue_asm(c->peer, buf);
			}
			else
			{
				sprintf(buf, "jp z,%s", sp = new_label());
				write_asm(buf);
				sprintf(buf, "%s:", sp);
				queue_asm(c->child, buf);
			}
		}
		else if (!strcmp(sword, "else"))
		{
			if (c->aq_base)
			{
				if (asm_code->asm_base)
				{
					c->aq_base->prev = asm_code->cur_asm;
					asm_code->cur_asm->next = c->aq_base;
				}
				else
					asm_code->asm_base = c->aq_base;
				asm_code->cur_asm = c->aq;
				c->aq_base = NULL;
			}
		}
		else if (!strcmp(sword, "do"))
		{
			if (!c->child)
			{
				PRINT_ERR("No statement following `do'");
				return;
			}
			if (!c->peer)
			{
				PRINT_ERR("No `while' following `do'");
				return;
			}
			if (strcmp(first_word(buf, c->peer->str), "while"))
			{
				PRINT_ERR("No `while' following `do'");
				return;
			}

			sprintf(buf, "%s:", sp = new_label());
			write_asm(buf);

			sprintf(buf, "%s:", sp3 = new_label());
			queue_asm(c->child, buf);

			queue_asm(c->peer, "pop8 a");
			queue_asm(c->peer, "test8 a");
			sprintf(buf, "jp nz,%s", sp);
			queue_asm(c->peer, buf);
			sprintf(buf, "%s:", sp2 = new_label());
			queue_asm(c->peer, buf);

			c->child->_continue = sp3;
			c->child->_break = sp2;
		}
		else if (!strcmp(sword, "while"))
		{
			flag = 0;
			if (c->pparent)
				if (!strcmp(c->pparent->str, "do"))
					flag = 1;

			if (flag)
			{
				eval(cp, &v, 0, 0);
			}
			else
			{
				if (!c->child)
				{
					PRINT_ERR("No statement following `while'");
					return;
				}

				sprintf(buf, "%s:", sp = new_label());
				write_asm(buf);
				eval(cp, &v, 0, 0);
				sprintf(buf, "pop%d a", v.load_size);
				write_asm(buf);
				sprintf(buf, "test%d a", v.load_size);
				write_asm(buf);
				sprintf(buf, "jp z,%s", sp2 = new_label());
				write_asm(buf);

				sprintf(buf, "jp %s", sp);
				queue_asm(c->child, buf);
				sprintf(buf, "%s:", sp2);
				queue_asm(c->child, buf);

				c->child->_continue = sp;
				c->child->_break = sp2;
			}
		}
		else if (!strcmp(sword, "for"))
		{
			trim_spaces(cp);
			for (sp3 = ++cp; *sp3; ++sp3)
			{
				if (quote)
				{
					if (slash)
						slash = 0;
					else if (*sp3 == '\\')
						slash = 1;
					else if (*sp3 == '"')
						quote = 0;
				}
				else if (squote)
				{
					if (slash)
						slash = 0;
					else if (*sp3 == '\\')
						slash = 1;
					else if (*sp3 == '\'')
						squote = 0;
				}
				else if (*sp3 == '"')
					quote = 1;
				else if (*sp3 == '\'')
					squote = 1;
				else if (*sp3 == '(' || *sp3 == '[')
					paren = 1;
				else if (*sp3 == ';')
					break;
			}

			if (!*sp3)
			{
				PRINT_ERR("Parse error in `for' expression list");
				return;
			}
			*sp3 = '\0';
			eval(cp, &v, 0, 0);
			write_asm("pop a");

			sprintf(buf, "%s:", sp = new_label());
			write_asm(buf);
			for (cp = ++sp3; *sp3; ++sp3)
			{
				if (quote)
				{
					if (slash)
						slash = 0;
					else if (*sp3 == '\\')
						slash = 1;
					else if (*sp3 == '"')
						quote = 0;
				}
				else if (squote)
				{
					if (slash)
						slash = 0;
					else if (*sp3 == '\\')
						slash = 1;
					else if (*sp3 == '\'')
						squote = 0;
				}
				else if (*sp3 == '"')
					quote = 1;
				else if (*sp3 == '\'')
					squote = 1;
				else if (*sp3 == '(' || *sp3 == '[')
					paren = 1;
				else if (*sp3 == ';')
					break;
			}

			if (!*sp3)
			{
				PRINT_ERR("Parse error in `for' expression list");
				return;
			}
			*sp3 = '\0';
			eval(cp, &v, 0, 0);
			sprintf(buf, "pop%d a", v.load_size);
			write_asm(buf);
			sprintf(buf, "test%d a", v.load_size);
			write_asm(buf);
			sprintf(buf, "jp z,%s", sp2 = new_label());
			write_asm(buf);

			cp = sp3+1;
			if (strlen(cp))
				cp[strlen(cp)-1] = '\0';

			c2 = code_end(c);
			c2->exprq = (char *)calloc(strlen(cp)+1,
				sizeof(char));
			strcpy(c2->exprq, cp);
			sprintf(buf, "jp %s", sp);
			queue_asm(c2, buf);
			sprintf(buf, "%s:", sp2);
			queue_asm(c2, buf);

			c->child->_continue = sp;
			c->child->_break = sp2;
		}
		else if (!strcmp(sword, "continue"))
		{
			trim_spaces(cp);
			if (cp[0])
			{
				PRINT_ERR("Parse error after `continue'");
				return;
			}
			if (!_continue)
			{
				PRINT_ERR("`continue' not within an appropriate block");
				return;
			}

			sprintf(buf, "jp %s", _continue);
			write_asm(buf);
		}
		else if (!strcmp(sword, "break"))
		{
			trim_spaces(cp);
			if (cp[0])
			{
				PRINT_ERR("Parse error after `break'");
				return;
			}
			if (!_break)
			{
				PRINT_ERR("`break' not within an appropriate block");
				return;
			}

			sprintf(buf, "jp %s", _break);
			write_asm(buf);
		}
		else if (!strcmp(sword, "goto"))
		{
			trim_spaces(cp);
			if (!cp[0])
			{
				PRINT_ERR("No label specified for `goto'");
				return;
			}
			if (!is_valid_id(cp))
			{
				sprintf(buf, "Invalid identifier `%s'", cp);
				PRINT_ERR(buf);
				return;
			}

			for (ul = func_base->label_base; ul; ul = ul->next)
				if (!strcmp(cp, ul->user))
				{
					sprintf(buf, "jp %s", ul->real);
					write_asm(buf);
					break;
				}

			if (!ul)
			{
				for (ul = func_base->orphan_base; ul; ul = ul->next)
					if (!strcmp(cp, ul->user))
					{
						sprintf(buf, "jp %s", ul->real);
						write_asm(buf);
						break;
					}

				if (!ul)
				{
					ul = (ulbl *)calloc(1, sizeof(ulbl));

					ul->user = (char *)calloc(strlen(cp)+1,
						sizeof(char));
					strcpy(ul->user, cp);
					ul->real = new_label();
					ul->pos = line_pos;
		
					ul->next = func_base->orphan_base;
					if (func_base->orphan_base)
						func_base->orphan_base->prev = ul;
					func_base->orphan_base = ul;
				}
			}
		}
		else if (!strcmp(sword, "switch"))
		{
			if (!c->child)
			{
				PRINT_ERR("No statement following `switch'");
				return;
			}

			eval(cp, &v, 0, 0);
			sprintf(buf, "pop%d a", v.load_size);
			write_asm(buf);

			def_label = NULL;
			case_base = NULL;
			switch_search(swbase = c->child);

			for (cs = case_base; cs; cs = cs2)
			{
				cs2 = cs->next;
				sprintf(buf, "ld%d b,%d", v.load_size, cs->val);
				write_asm(buf);
				sprintf(buf, "%scp%d a,b", (v.attrs & SIGNED) ?
					"s" : "", v.load_size);
				write_asm(buf);
				sprintf(buf, "jp z,%s", cs->label);
				write_asm(buf);
				free(cs);
			}

			sp = new_label();
			if (def_label)
				sprintf(buf, "jp %s", def_label);
			else
				sprintf(buf, "jp %s", sp);
			write_asm(buf);

			sprintf(buf, "%s:", sp);
			queue_asm(c->child, buf);

			c->child->_break = sp;
		}
		else if (!strcmp(sword, "return"))
		{
			trim_spaces(cp);

			if (cp[0])
			{
				eval(cp, &v, 0, 0);
				if (func_base->code_base->size)
					write_asm("lfree");
				write_asm("ret");
			}
			else
			{
				if (func_base->code_base->size)
					write_asm("lfree");
				write_asm("ld a,0");
				write_asm("retv a");
			}
		}
		else
		{
			eval(c->str, &v, 0, 0);
			write_asm("pop a");
		}
		if (c->child)
			do_statements(c->child, parent, c->child->_break ?
				c->child->_break : _break, c->child->_continue ?
				c->child->_continue : _continue);
		if (c->peer)
			do_statements(c->peer, parent, c->peer->_break ?
				c->peer->_break : _break, c->peer->_continue ?
				c->peer->_continue : _continue);

		break;
	}

	if (c->exprq)
	{
		eval(c->exprq, &v, 0, 0);
		write_asm("pop a");
	}
	if (c->aq_base)
	{
		if (asm_code->asm_base)
		{
			c->aq_base->prev = asm_code->cur_asm;
			asm_code->cur_asm->next = c->aq_base;
		}
		else
			asm_code->asm_base = c->aq_base;
		asm_code->cur_asm = c->aq;
	}
}

void cleanup()
{
	file_info *fp, *fp2;
	macro *m, *m2;
	func *f, *f2;

	if (finfo.name)
	{
		free(finfo.name);
		free(finfo.text);
		free(finfo.lines);
	}
	if (finfo2.name)
	{
		free(finfo2.text);
		free(finfo2.lines);
	}

	for (fp = flist_base; fp; fp = fp2)
	{
		fp2 = fp->next;
		free(fp->name);
		if (fp->text)
			free(fp->text);
		free(fp->lines);
		free(fp);
	}
	for (m = macro_base; m; m = m2)
	{
		m2 = m->next;
		free(m->name);
		free(m->value);
		free_lstr(m->args);
		free(m);
	}
	for (f = func_base; f; f = f2)
	{
		f2 = f->next;
		free_var(&f->type, 0);
		free_ulbl(f->label_base);
		free_ulbl(f->orphan_base);
		free(f);
	}

	free_lstr(label_base);
	if (global.code_base)
		free_code(global.code_base);
}

void print_code(code *c)
{
	for (; c; c = c->next)
	{
		if (c->child)
		{
			puts("[");
			print_code(c->child);
			puts("]");
		}
		else
			print_err(c->file, c->pos, c->str);
	}
}

main(int argc, char *argv[])
{
	char source_path[PATH_LENGTH], buf[WORD_LENGTH];
	char fname[PATH_LENGTH], ofname[PATH_LENGTH] = "out.xc";
	int j, osz, pos = 0, line = 0, next_line;
	time_t tim;
	struct tm *theTime;
	char *sp, amode = 0;
	func *f;
	var *v;
	asmt *a;
	macro *m;
	FILE *fp;

	memset(&settings, 0, sizeof(settings));
	finfo.name = NULL;
	finfo2.name = NULL;

	if (argc == 1)
	{
		puts("xcc xCode C compiler");
		puts("Copyright Adam Chlipala 1999 (psion@tpu.org)");
		puts("Usage: xcc [flags] <file.c> [flags]\n");
		puts("Flags:");
		puts("-o <filename>: Set module output file name");
		puts("        -c/-S: Skip final assembly stage and just leave out.xas assembly source");
		puts("        -ansi: Enable/disable features to meet ANSI C standard");
		puts("    -save_asm: Compile as normal, but leave out.xas as well");
		puts("       -nonet: Compile module without network streaming support");
		return 0;
	}

	m = (macro *)calloc(1, sizeof(macro));
	strcpy(buf, "__XCC__");
	m->name = (char *)malloc(strlen(buf)+1);
	strcpy(m->name, buf);
	m->value = (char *)malloc(1);
	m->value[0] = '\0';
	macro_base = m;

	for (j = 1; j < argc; ++j)
	{
		if (!amode && argv[j][0] == '-')
		{
			if (!strcmp(argv[j], "-ansi"))
			{
				settings.ansi = settings.ansi_on = 1;
				m = (macro *)calloc(1, sizeof(macro));
				strcpy(buf, "__STDC__");
				m->name = (char *)malloc(strlen(buf)+1);
				strcpy(m->name, buf);
				m->value = (char *)malloc(1);
				m->value[0] = '\0';
				m->next = macro_base;
				macro_base = m;
			}
			else if (!strcmp(argv[j], "-o"))
				amode = 1;
			else if (!strcmp(argv[j], "-c") || !strcmp(argv[j], "-S"))
				settings.skip_asm = settings.save_asm = 1;
			else if (!strcmp(argv[j], "-save_asm"))
				settings.save_asm = 1;
			else if (!strcmp(argv[j], "-nonet"))
				settings.nonet = 1;
			else if (!strcmp(argv[j], "-ht"))
				settings.host_table = 1;
		}
		else
		{
			switch (amode)
			{
			case 0:
				strcpy(fname, argv[j]);
				break;
			case 1:
				strcpy(ofname, argv[j]);
				break;
			}
			amode = 0;
		}
	}
	if (amode)
	{
		puts("Argument not supplied for last flag");
		return 1;
	}

	time(&tim);
	theTime = localtime(&tim);
	sprintf(date_str, "%3s %02d %04d", months[theTime->tm_mon], 
		theTime->tm_mday, 1900 + theTime->tm_year);
	sprintf(time_str, "%02d:%02d:%02d", theTime->tm_hour,
		theTime->tm_min, theTime->tm_sec);

	getcwd(path, PATH_LENGTH);
	sprintf(source_path, "%s/%s", path, fname);
	standardize_path(source_path);

	if (open_file(&finfo, source_path, fname))
	{
		printf("Error opening %s\n", argv[1]);
		return 1;
	}

	atexit(cleanup);

	osz = BLOCK_SIZE;
	finfo2.text = (char *)calloc(BLOCK_SIZE, sizeof(char));
	finfo2.n_lines = 0;
	finfo2.lines = (line_info *)malloc(0);

	if (preprocess(&finfo2, &finfo, finfo.text, &pos, &osz)
		|| errcount)
		return 1;
//	puts(finfo2.text);
	finfo2.text[pos] = '\0';
	finfo2.size = ++pos;
	if (errcount)
		return 1;

	pos = 0;
	cur_file = finfo2.lines[0].file;
	cur_line = finfo2.lines[0].line;
	if (finfo2.n_lines >= 2)
		next_line = finfo2.lines[1].pos;
	else
		next_line = -1;

	memset(&global, 0, sizeof(func));
	cur_func = &global;

	sp = finfo2.text;
	global.code_base = (code *)calloc(1, sizeof(code));
	compile(&finfo2, &sp, &pos, &line, &next_line, global.code_base);
	free_lstr(temp_td_base);
	if (errcount)
		return 1;

	global.code_base->type = MULTI;
	asm_code = global.code_base;
	global.code_base->size = global.code_base->self_size = 1;
	do_statements(global.code_base, NULL, NULL, NULL);

	if (errcount)
		return 1;

	if (!(fp = fopen("out.xas", "w")))
	{
		puts("Error opening output file");
		return 1;
	}

	fprintf(fp, "mbase %d\n", global.code_base->self_size);
	for (a = global.code_base->asm_base; a; a = a->next)
		fprintf(fp, "%s\n", a->str);
	fputs("call _main\n", fp);
//	if (!settings.nonet)
//		fputs("imode\nrmode\nsync\nemode\nemode\nomode\nrmode\nsync\nemode\nemode\n", fp);
	fputs("ret\n", fp);
	for (f = func_base; f; f = f->next)
	{
		if (!f->code_base)
			continue;
		fprintf(fp, "\n_%s:\n", f->type.name);
		if (f->code_base->size)
			fprintf(fp, "lalloc %d\n", f->code_base->size);
		for (v = f->type.args; v; v = v->next_arg)
		{
			fprintf(fp, "pop%d a\n", v->load_size);
			fprintf(fp, "laddr %d\npop b\n", v->loc);
			fprintf(fp, "ld%d (b),a\n", v->load_size);
		}
		for (a = f->code_base->asm_base; a; a = a->next)
			fprintf(fp, "%s\n", a->str);
		if (f->code_base->size)
			fputs("lfree\n", fp);
		fputs("ld a,0\nretv a\n", fp);
	}
	fclose(fp);

	if (settings.skip_asm)
		return 0;

	sprintf(buf, "xasm out.xas %s", ofname);
	if (system(buf))
		return 1;

	if (!settings.save_asm)
		remove("out.xas");

	return 0;
}